from ._base import RuleMapping


def create_classic_to_core_mapping() -> RuleMapping:
    return RuleMapping.model_validate(_CLASSIC_TO_CORE)


_CLASSIC_TO_CORE = {
    "classes": [
        {"destination": {"Class": "cdf_cdm:CogniteAsset"}, "source": {"Class": "classic:Asset"}},
        {"destination": {"Class": "cdf_cdm:DataProduct"}, "source": {"Class": "classic:DataSet"}},
        {"destination": {"Class": "cdf_cdm:CogniteActivity"}, "source": {"Class": "classic:Event"}},
        {"destination": {"Class": "cdf_cdm:CogniteFile"}, "source": {"Class": "classic:File"}},
        {"destination": {"Class": "cdf_cdm:CogniteTimeSeries"}, "source": {"Class": "classic:TimeSeries"}},
    ],
    "properties": [
        {
            "destination": {"Class": "cdf_cdm:CogniteAsset", "Property": "externalId"},
            "source": {"Class": "classic:Asset", "Property": "externalId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteAsset", "Property": "name"},
            "source": {"Class": "classic:Asset", "Property": "name"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteAsset", "Property": "parent"},
            "source": {"Class": "classic:Asset", "Property": "parentId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteAsset", "Property": "description"},
            "source": {"Class": "classic:Asset", "Property": "description"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteSourceSystem", "Property": "name"},
            "source": {"Class": "classic:Asset", "Property": "source"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteAsset", "Property": "tag"},
            "source": {"Class": "classic:Asset", "Property": "labels"},
        },
        {
            "destination": {"Class": "cdf_cdm:DataProduct", "Property": "externalId"},
            "source": {"Class": "classic:DataSet", "Property": "externalId"},
        },
        {
            "destination": {"Class": "cdf_cdm:DataProduct", "Property": "name"},
            "source": {"Class": "classic:DataSet", "Property": "name"},
        },
        {
            "destination": {"Class": "cdf_cdm:DataProduct", "Property": "description"},
            "source": {"Class": "classic:DataSet", "Property": "description"},
        },
        {
            "destination": {"Class": "cdf_cdm:DataProduct", "Property": "metadata"},
            "source": {"Class": "classic:DataSet", "Property": "metadata"},
        },
        {
            "destination": {"Class": "cdf_cdm:DataProduct", "Property": "writeProtected"},
            "source": {"Class": "classic:DataSet", "Property": "writeProtected"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteActivity", "Property": "externalId"},
            "source": {"Class": "classic:Event", "Property": "externalId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteSchedulable", "Property": "startTime"},
            "source": {"Class": "classic:Event", "Property": "startTime"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteSchedulable", "Property": "endTime"},
            "source": {"Class": "classic:Event", "Property": "endTime"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteActivity", "Property": "description"},
            "source": {"Class": "classic:Event", "Property": "description"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteActivity", "Property": "assets"},
            "source": {"Class": "classic:Event", "Property": "assetIds"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteSourceSystem", "Property": "name"},
            "source": {"Class": "classic:Event", "Property": "source"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "externalId"},
            "source": {"Class": "classic:File", "Property": "externalId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "name"},
            "source": {"Class": "classic:File", "Property": "name"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "directory"},
            "source": {"Class": "classic:File", "Property": "directory"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteSourceSystem", "Property": "name"},
            "source": {"Class": "classic:File", "Property": "source"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "mimeType"},
            "source": {"Class": "classic:File", "Property": "mimeType"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "assets"},
            "source": {"Class": "classic:File", "Property": "assetIds"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "sourceCreatedTime"},
            "source": {"Class": "classic:File", "Property": "sourceCreatedTime"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteFile", "Property": "sourceUpdatedTime"},
            "source": {"Class": "classic:File", "Property": "sourceModifiedTime"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "externalId"},
            "source": {"Class": "classic:TimeSeries", "Property": "externalId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "name"},
            "source": {"Class": "classic:TimeSeries", "Property": "name"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "type"},
            "source": {"Class": "classic:TimeSeries", "Property": "isString"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "sourceUnit"},
            "source": {"Class": "classic:TimeSeries", "Property": "unit"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "unit"},
            "source": {"Class": "classic:TimeSeries", "Property": "unitExternalId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "assets"},
            "source": {"Class": "classic:TimeSeries", "Property": "assetId"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "isStep"},
            "source": {"Class": "classic:TimeSeries", "Property": "isStep"},
        },
        {
            "destination": {"Class": "cdf_cdm:CogniteTimeSeries", "Property": "description"},
            "source": {"Class": "classic:TimeSeries", "Property": "description"},
        },
    ],
}



import warnings
from typing import cast

from rdflib import RDF, Graph, Literal, Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._constants import CLASSIC_CDF_NAMESPACE, DEFAULT_NAMESPACE
from cognite.neat._graph import extractors
from cognite.neat._issues.warnings import ResourceNotFoundWarning
from cognite.neat._utils.rdf_ import remove_namespace_from_uri

from ._base import BaseTransformer


class AddAssetDepth(BaseTransformer):
    description: str = "Adds depth of asset in the asset hierarchy to the graph"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__)})

    _parent_template: str = """SELECT ?child ?parent WHERE {{
                              <{asset_id}> <{parent_prop}> ?child .
                              OPTIONAL{{?child <{parent_prop}>+ ?parent .}}}}"""

    _root_template: str = """SELECT ?root WHERE {{
                             <{asset_id}> <{root_prop}> ?root .}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        root_prop: URIRef | None = None,
        parent_prop: URIRef | None = None,
        depth_typing: dict[int, str] | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.root_prop = root_prop or DEFAULT_NAMESPACE.root
        self.parent_prop = parent_prop or DEFAULT_NAMESPACE.parent
        self.depth_typing = depth_typing

    def transform(self, graph: Graph) -> None:
        """Adds depth of asset in the asset hierarchy to the graph."""
        for result in graph.query(f"SELECT DISTINCT ?asset_id WHERE {{?asset_id a <{self.asset_type}>}}"):
            asset_id = cast(tuple, result)[0]
            if depth := self.get_depth(graph, asset_id, self.root_prop, self.parent_prop):
                graph.add((asset_id, DEFAULT_NAMESPACE.depth, Literal(depth)))

                if self.depth_typing and (type_ := self.depth_typing.get(depth, None)):
                    # remove existing type
                    graph.remove((asset_id, RDF.type, None))

                    # add new type
                    graph.add((asset_id, RDF.type, DEFAULT_NAMESPACE[type_]))

    @classmethod
    def get_depth(
        cls,
        graph: Graph,
        asset_id: URIRef,
        root_prop: URIRef,
        parent_prop: URIRef,
    ) -> int | None:
        """Get asset depth in the asset hierarchy."""

        # Handles non-root assets
        if result := list(graph.query(cls._parent_template.format(asset_id=asset_id, parent_prop=parent_prop))):
            return len(cast(list[tuple], result)) + 2 if cast(list[tuple], result)[0][1] else 2

        # Handles root assets
        elif (
            (result := list(graph.query(cls._root_template.format(asset_id=asset_id, root_prop=root_prop))))
            and len(cast(list[tuple], result)) == 1
            and cast(list[tuple], result)[0][0] == asset_id
        ):
            return 1
        else:
            return None


class AssetTimeSeriesConnector(BaseTransformer):
    description: str = "Connects assets to timeseries, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.TimeSeriesExtractor.__name__),
        }
    )
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{timeseries_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        timeseries_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.timeseries_type = timeseries_type or DEFAULT_NAMESPACE.TimeSeries
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for ts_id_result in graph.query(
            f"SELECT DISTINCT ?timeseries_id WHERE {{?timeseries_id a <{self.timeseries_type}>}}"
        ):
            timeseries_id: URIRef = cast(tuple, ts_id_result)[0]

            if asset_id_res := list(
                graph.query(
                    self._asset_template.format(
                        timeseries_id=timeseries_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # timeseries can be connected to only one asset in the graph
                asset_id = cast(list[tuple], asset_id_res)[0][0]
                graph.add((asset_id, DEFAULT_NAMESPACE.timeSeries, timeseries_id))


class AssetSequenceConnector(BaseTransformer):
    description: str = "Connects assets to sequences, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.SequencesExtractor.__name__),
        }
    )
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{sequence_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        sequence_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.sequence_type = sequence_type or DEFAULT_NAMESPACE.Sequence
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for sequency_id_result in graph.query(
            f"SELECT DISTINCT ?sequence_id WHERE {{?sequence_id a <{self.sequence_type}>}}"
        ):
            sequence_id: URIRef = cast(tuple, sequency_id_result)[0]

            if asset_id_res := list(
                graph.query(
                    self._asset_template.format(
                        sequence_id=sequence_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # sequence can be connected to only one asset in the graph
                asset_id = cast(list[tuple], asset_id_res)[0][0]
                graph.add((asset_id, DEFAULT_NAMESPACE.sequence, sequence_id))


class AssetFileConnector(BaseTransformer):
    description: str = "Connects assets to files, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.FilesExtractor.__name__),
        }
    )
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{file_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        file_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.file_type = file_type or DEFAULT_NAMESPACE.File
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for sequency_id_result in graph.query(f"SELECT DISTINCT ?file_id WHERE {{?file_id a <{self.file_type}>}}"):
            file_id: URIRef = cast(tuple, sequency_id_result)[0]

            if assets_id_res := list(
                graph.query(
                    self._asset_template.format(
                        file_id=file_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # files can be connected to multiple assets in the graph
                for (asset_id,) in cast(list[tuple], assets_id_res):
                    graph.add((asset_id, DEFAULT_NAMESPACE.file, file_id))


class AssetEventConnector(BaseTransformer):
    description: str = "Connects assets to events, thus forming bi-directional connection"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.EventsExtractor.__name__),
        }
    )
    _asset_template: str = """SELECT ?asset_id WHERE {{
                              <{event_id}> <{asset_prop}> ?asset_id .
                              ?asset_id a <{asset_type}>}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        event_type: URIRef | None = None,
        asset_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.event_type = event_type or DEFAULT_NAMESPACE.Event
        self.asset_prop = asset_prop or DEFAULT_NAMESPACE.asset

    def transform(self, graph: Graph) -> None:
        for event_id_result in graph.query(f"SELECT DISTINCT ?event_id WHERE {{?event_id a <{self.event_type}>}}"):
            event_id: URIRef = cast(tuple, event_id_result)[0]

            if assets_id_res := list(
                graph.query(
                    self._asset_template.format(
                        event_id=event_id,
                        asset_prop=self.asset_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # files can be connected to multiple assets in the graph
                for (asset_id,) in cast(list[tuple], assets_id_res):
                    graph.add((asset_id, DEFAULT_NAMESPACE.event, event_id))


class AssetRelationshipConnector(BaseTransformer):
    description: str = "Connects assets via relationships"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.RelationshipsExtractor.__name__),
        }
    )
    _asset_template: str = """SELECT ?source ?target WHERE {{
                              <{relationship_id}> <{relationship_source_xid_prop}> ?source_xid .
                              ?source <{asset_xid_property}> ?source_xid .
                              ?source a <{asset_type}> .

                              <{relationship_id}> <{relationship_target_xid_prop}> ?target_xid .
                              ?target <{asset_xid_property}> ?target_xid .
                              ?target a <{asset_type}> .}}"""

    def __init__(
        self,
        asset_type: URIRef | None = None,
        relationship_type: URIRef | None = None,
        relationship_source_xid_prop: URIRef | None = None,
        relationship_target_xid_prop: URIRef | None = None,
        asset_xid_property: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.relationship_type = relationship_type or DEFAULT_NAMESPACE.Relationship
        self.relationship_source_xid_prop = relationship_source_xid_prop or DEFAULT_NAMESPACE.source_external_id
        self.relationship_target_xid_prop = relationship_target_xid_prop or DEFAULT_NAMESPACE.target_external_id
        self.asset_xid_property = asset_xid_property or DEFAULT_NAMESPACE.external_id

    def transform(self, graph: Graph) -> None:
        for relationship_id_result in graph.query(
            f"SELECT DISTINCT ?relationship_id WHERE {{?relationship_id a <{self.relationship_type}>}}"
        ):
            relationship_id: URIRef = cast(tuple, relationship_id_result)[0]

            if assets_id_res := list(
                graph.query(
                    self._asset_template.format(
                        relationship_id=relationship_id,
                        asset_xid_property=self.asset_xid_property,
                        relationship_source_xid_prop=self.relationship_source_xid_prop,
                        relationship_target_xid_prop=self.relationship_target_xid_prop,
                        asset_type=self.asset_type,
                    )
                )
            ):
                # files can be connected to multiple assets in the graph
                for source_asset_id, target_asset_id in cast(list[tuple], assets_id_res):
                    # create a relationship between the two assets
                    graph.add(
                        (
                            source_asset_id,
                            DEFAULT_NAMESPACE.relationship,
                            relationship_id,
                        )
                    )
                    graph.add(
                        (
                            target_asset_id,
                            DEFAULT_NAMESPACE.relationship,
                            relationship_id,
                        )
                    )

                    # add source and target to the relationship
                    graph.add((relationship_id, DEFAULT_NAMESPACE.source, source_asset_id))
                    graph.add((relationship_id, DEFAULT_NAMESPACE.target, target_asset_id))

                    # remove properties that are not needed, specifically the external ids
                    graph.remove((relationship_id, self.relationship_source_xid_prop, None))
                    graph.remove((relationship_id, self.relationship_target_xid_prop, None))


class RelationshipToSchemaTransformer(BaseTransformer):
    """Replaces relationships with a schema.

    This transformer analyzes the relationships in the graph and modifies them to be part of the schema
    for Assets, Events, Files, Sequences, and TimeSeries. Relationships without any properties
    are replaced by a simple relationship between the source and target nodes. Relationships with
    properties are replaced by a schema that contains the properties as attributes.

    Args:
        limit: The minimum number of relationships that need to be present for it
            to be converted into a schema. Default is 1.

    """

    def __init__(self, limit: int = 1, namespace: Namespace = CLASSIC_CDF_NAMESPACE) -> None:
        self._limit = limit
        self._namespace = namespace

    _NOT_PROPERTIES: frozenset[str] = frozenset(
        {"source_external_id", "target_external_id", "external_id", "source_type", "target_type"}
    )
    _RELATIONSHIP_NODE_TYPES: tuple[str, ...] = tuple(["Asset", "Event", "File", "Sequence", "TimeSeries"])
    description = "Replaces relationships with a schema"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.RelationshipsExtractor.__name__)})

    _count_by_source_target = """PREFIX classic: <{namespace}>

SELECT (COUNT(?instance) AS ?instanceCount)
WHERE {{
  ?instance a classic:Relationship .
  ?instance classic:source_type classic:{source_type} .
  ?instance classic:target_type classic:{target_type} .
}}"""

    _instances = """PREFIX classic: <{namespace}>

SELECT ?instance
WHERE {{
    ?instance a classic:Relationship .
    ?instance classic:source_type classic:{source_type} .
    ?instance classic:target_type classic:{target_type} .
}}"""
    _lookup_entity_query = """PREFIX classic: <{namespace}>

SELECT ?entity
WHERE {{
    ?entity a classic:{entity_type} .
    ?entity classic:external_id "{external_id}" .
}}"""

    def transform(self, graph: Graph) -> None:
        for source_type in self._RELATIONSHIP_NODE_TYPES:
            for target_type in self._RELATIONSHIP_NODE_TYPES:
                query = self._count_by_source_target.format(
                    namespace=self._namespace, source_type=source_type, target_type=target_type
                )
                for instance_count in graph.query(query):
                    if int(instance_count[0]) < self._limit:  # type: ignore[index, arg-type]
                        continue
                    query = self._instances.format(
                        namespace=self._namespace, source_type=source_type, target_type=target_type
                    )
                    for result in graph.query(query):
                        instance_id = cast(URIRef, result[0])  # type: ignore[index, misc]
                        self._convert_relationship_to_schema(graph, instance_id, source_type, target_type)

    def _convert_relationship_to_schema(
        self, graph: Graph, instance_id: URIRef, source_type: str, target_type: str
    ) -> None:
        result = cast(list[ResultRow], list(graph.query(f"DESCRIBE <{instance_id}>")))
        object_by_predicates = cast(
            dict[str, URIRef | Literal], {remove_namespace_from_uri(row[1]): row[2] for row in result}
        )
        source_external_id = cast(URIRef, object_by_predicates["source_external_id"])
        target_source_id = cast(URIRef, object_by_predicates["target_external_id"])
        try:
            source_id = self._lookup_entity(graph, source_type, source_external_id)
        except ValueError:
            warnings.warn(ResourceNotFoundWarning(source_external_id, "class", str(instance_id), "class"), stacklevel=2)
            return None
        try:
            target_id = self._lookup_entity(graph, target_type, target_source_id)
        except ValueError:
            warnings.warn(ResourceNotFoundWarning(target_source_id, "class", str(instance_id), "class"), stacklevel=2)
            return None
        external_id = str(object_by_predicates["external_id"])
        # If there is properties on the relationship, we create a new intermediate node
        self._create_node(graph, object_by_predicates, external_id, source_id, target_id, self._predicate(target_type))

        for triple in result:
            graph.remove(triple)  # type: ignore[arg-type]

    def _lookup_entity(self, graph: Graph, entity_type: str, external_id: str) -> URIRef:
        query = self._lookup_entity_query.format(
            namespace=self._namespace, entity_type=entity_type, external_id=external_id
        )
        result = list(graph.query(query))
        if len(result) == 1:
            return cast(URIRef, result[0][0])  # type: ignore[index]
        raise ValueError(f"Could not find entity with external_id {external_id} and type {entity_type}")

    def _create_node(
        self,
        graph: Graph,
        objects_by_predicates: dict[str, URIRef | Literal],
        external_id: str,
        source_id: URIRef,
        target_id: URIRef,
        predicate: URIRef,
    ) -> None:
        """Creates a new intermediate node for the relationship with properties."""
        # Create new node
        instance_id = self._namespace[external_id]
        graph.add((instance_id, RDF.type, self._namespace["Edge"]))
        for prop_name, object_ in objects_by_predicates.items():
            if prop_name in self._NOT_PROPERTIES:
                continue
            graph.add((instance_id, self._namespace[prop_name], object_))

        # Connect the new node to the source and target nodes
        graph.add((source_id, predicate, instance_id))
        graph.add((instance_id, self._namespace["end_node"], target_id))

    def _predicate(self, target_type: str) -> URIRef:
        return self._namespace[f"relationship{target_type.capitalize()}"]



from typing import cast

from rdflib import Graph

from cognite.neat._rules.importers._rdf._shared import (
    clean_up_properties,
    make_properties_compliant,
    parse_raw_properties_dataframe,
)


def parse_imf_to_properties(graph: Graph, language: str = "en") -> list[dict]:
    """Parse IMF elements from RDF-graph and extract properties to pandas dataframe.

    Args:
        graph: Graph containing imf elements
        language: Language to use for parsing, by default "en"

    Returns:
        List of dictionaries containing properties extracted from IMF elements

    !!! note "IMF Compliance"
        The IMF elements are expressed in RDF, primarily using SHACL and OWL. To ensure
        that the resulting properties are compliant with CDF, similar validation checks
        as in the OWL ontology importer are applied.

        For the IMF-types more of the compliance logic is placed directly in the SPARQL
        query. Among these are the creation of class and property names not starting
        with a number, ensuring property types as well as default cardinality boundraries.

        IMF-attributes are considered both classes and properties. This kind of punning
        is necessary to capture additional information carried by attributes. They carry,
        among other things, a set of relationsships to reference terms, units of measure,
        and qualifiers that together make up the meaning of the attribute. These references
        are listed as additional properties with default values.
    """

    query = """
    SELECT DISTINCT ?class ?property ?name ?description ?valueType ?minCount ?maxCount ?default ?reference
    ?match ?comment ?propertyType
    WHERE
    {
        # Finding IMF-blocks and terminals
        {
            VALUES ?classType { imf:BlockType imf:TerminalType }
            ?imfClass a ?classType ;
                sh:property ?propertyShape .
                ?propertyShape sh:path ?imfProperty .

            OPTIONAL { ?imfProperty skos:prefLabel ?name . }
            OPTIONAL { ?imfProperty skos:definition | skos:description ?description . }
            OPTIONAL { ?imfProperty rdfs:range ?range . }
            OPTIONAL { ?imfProperty a ?type . }
            OPTIONAL { ?propertyShape sh:minCount ?minCardinality} .
            OPTIONAL { ?propertyShape sh:maxCount ?maxCardinality} .
            OPTIONAL { ?propertyShape sh:hasValue ?defualt . }
            OPTIONAL { ?propertyShape sh:class | sh:qualifiedValueShape/sh:class ?valueShape .}
        }
        UNION
        # Finding the IMF-attribute types
        {
            ?imfClass a imf:AttributeType ;
                ?imfProperty ?default .

            # The following information is used to describe the attribute when it is connected to a block or a terminal
            # and not duplicated here.
            # Note: Bug in PCA has lead to the use non-existing term skos:description. This will be replaced
            # with the correct skos:definition in the near future, so both terms are included here.
            FILTER(?imfProperty != rdf:type && ?imfProperty != skos:prefLabel &&
                ?imfProperty != skos:defintion && ?imfProperty != skos:description)
        }

        # Finding the last segment of the class IRI
        BIND(STR(?imfClass) AS ?classString)
        BIND(REPLACE(?classString, "^.*[/#]([^/#]*)$", "$1") AS ?tempClassSegment)
        BIND(REPLACE(?tempClassSegment, "-", "_") AS ?classSegment)
        BIND(IF(CONTAINS(?classString, "imf/"), CONCAT("IMF_", ?classSegment) , ?classSegment) AS ?class)


        # Finding the last segment of the property IRI
        BIND(STR(?imfProperty) AS ?propertyString)
        BIND(REPLACE(?propertyString, "^.*[/#]([^/#]*)$", "$1") AS ?tempPropertySegment)
        BIND(REPLACE(?tempPropertySegment, "-", "_") AS ?propertySegment)
        BIND(IF(CONTAINS(?propertyString, "imf/"), CONCAT("IMF_", ?propertySegment) , ?propertySegment) AS ?property)

        # Set the value type for the property based on sh:class, sh:qualifiedValueType or rdfs:range
        BIND(IF(BOUND(?valueShape), ?valueShape, IF(BOUND(?range) , ?range , ?valueShape)) AS ?valueIriType)

        # Finding the last segment of value types
        BIND(STR(?valueIriType) AS ?valueTypeString)
        BIND(REPLACE(?valueTypeString, "^.*[/#]([^/#]*)$", "$1") AS ?tempValueTypeSegment)
        BIND(REPLACE(?tempValueTypeSegment, "-", "_") AS ?valueTypeSegment)
        BIND(IF(CONTAINS(?valueTypeString, "imf/"), CONCAT("IMF_", ?valueTypeSegment) , ?valueTypeSegment)
            AS ?valueType)

        # Helper variable to set owl datatype- or object-property if this is not already set.
        BIND(IF( EXISTS {?imfProperty a ?tempPropertyType . FILTER(?tempPropertyType = owl:DatatypeProperty) },
            owl:DatatypeProperty,
            owl:ObjectProperty
            )
        AS ?propertyType)

        # Assert cardinality values if they do not exist
        BIND(IF(BOUND(?minCardinality), ?minCardinality, 0) AS ?minCount)
        BIND(IF(BOUND(?maxCardinality), ?maxCardinality, 1) AS ?maxCount)

        # Rebind the IRI of the IMF-attribute to the ?reference variable to align with dataframe column headers
        # This is solely for readability, the ?imfClass could have been returnered directly instead of ?reference
        BIND(?imfProperty AS ?reference)

        FILTER (!isBlank(?property))
        FILTER (!bound(?class) || !isBlank(?class))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
    }
    """

    raw_df = parse_raw_properties_dataframe(cast(list[tuple], list(graph.query(query.replace("en", language)))))
    if raw_df.empty:
        return []

    # group values and clean up
    processed_df = clean_up_properties(raw_df)

    # make compliant
    processed_df = make_properties_compliant(processed_df, importer="IMF")

    # drop column _property_type, which was a helper column:
    processed_df.drop(columns=["_property_type"], inplace=True)

    return processed_df.to_dict(orient="records")



import functools
from collections.abc import Callable
from typing import Any

try:
    from rich import print

    _PREFIX = "[bold red][ERROR][/bold red]"
except ImportError:
    _PREFIX = "[ERROR]"


class NeatSessionError(Exception):
    """Base class for all exceptions raised by the NeatSession class."""

    ...


def _intercept_session_exceptions(func: Callable):
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any):
        try:
            return func(*args, **kwargs)
        except NeatSessionError as e:
            action = func.__name__
            if action == "__call__":
                action = func.__qualname__.removesuffix(".__call__").removesuffix("API")
            print(f"{_PREFIX} Cannot {action}: {e}")

    return wrapper


def intercept_session_exceptions(cls: type):
    to_check = [cls]
    while to_check:
        cls = to_check.pop()
        for attr_name in dir(cls):
            if attr_name.startswith("_") and not attr_name == "__call__":
                continue
            attr = getattr(cls, attr_name)
            if callable(attr):
                setattr(cls, attr_name, _intercept_session_exceptions(attr))
            elif isinstance(attr, type):
                to_check.append(attr)
    return cls



from dataclasses import dataclass
from typing import Generic

from cognite.neat._issues._base import NeatWarning, ResourceType, T_Identifier, T_ReferenceIdentifier


# Name ResourceNeatWarning to avoid conflicts with the built-in ResourceWarning
@dataclass(unsafe_hash=True)
class ResourceNeatWarning(NeatWarning, Generic[T_Identifier]):
    """Base class for resource warnings {resource_type} with identifier {identifier}"""

    identifier: T_Identifier
    resource_type: ResourceType


@dataclass(unsafe_hash=True)
class ResourceNotFoundWarning(ResourceNeatWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier} referred by {referred_type} {referred_by} does not exist.
    This will be ignored."""

    fix = "Create the {resource_type}"

    referred_by: T_ReferenceIdentifier
    referred_type: str


@dataclass(unsafe_hash=True)
class ResourcesDuplicatedWarning(NeatWarning, Generic[T_Identifier]):
    """Duplicated {resource_type} with identifiers {resources} were found. {default_action}"""

    fix = "Remove the duplicate resources"

    resources: frozenset[T_Identifier]
    resource_type: ResourceType
    default_action: str


@dataclass(unsafe_hash=True)
class ResourceRetrievalWarning(NeatWarning, Generic[T_Identifier]):
    """Failed to retrieve {resource_type} with identifiers {resources}. Continuing without
    these resources."""

    extra = "The error was: {error}"

    fix = "Check the error."

    resources: frozenset[T_Identifier]
    resource_type: ResourceType
    error: str | None = None


class ResourceTypeNotSupportedWarning(ResourceNeatWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} is not supported. This will be ignored."""

    resource_type: str  # type: ignore[assignment]



from collections.abc import Iterable
from pathlib import Path
from typing import cast

from rdflib import URIRef
from rdflib.util import guess_format

from cognite.neat._constants import DEFAULT_BASE_URI
from cognite.neat._graph._shared import rdflib_to_mime_types
from cognite.neat._graph.extractors._base import BaseExtractor
from cognite.neat._graph.models import Triple
from cognite.neat._issues._base import IssueList
from cognite.neat._issues.errors import FileNotFoundNeatError, FileTypeUnexpectedError


class RdfFileExtractor(BaseExtractor):
    """Extract data from RDF files into Neat.

    Args:
        filepath (Path): The path to the RDF file.
        mime_type (MIMETypes, optional): The MIME type of the RDF file. Defaults to "application/rdf+xml".
        base_uri (URIRef, optional): The base URI to use. Defaults to None.
    """

    def __init__(
        self,
        filepath: Path,
        base_uri: URIRef = DEFAULT_BASE_URI,
        issue_list: IssueList | None = None,
    ):
        self.issue_list = issue_list or IssueList(title=f"{filepath.name}")

        self.filepath = filepath
        self.mime_type = rdflib_to_mime_types(cast(str, guess_format(str(self.filepath))))
        self.base_uri = base_uri

        if not self.filepath.exists():
            self.issue_list.append(FileNotFoundNeatError(self.filepath))

        if not self.mime_type:
            self.issue_list.append(
                FileTypeUnexpectedError(
                    self.filepath,
                    frozenset([".rdf", ".ttl", ".nt", ".n3", ".owl", ".nq", ".trig"]),
                )
            )

    def extract(self) -> Iterable[Triple]:
        raise NotImplementedError()



from rdflib import XSD, Graph, URIRef

from cognite.neat._constants import UNKNOWN_TYPE
from cognite.neat._graph.queries import Queries
from cognite.neat._utils.rdf_ import remove_namespace_from_uri

from ._base import BaseTransformer


class SplitMultiValueProperty(BaseTransformer):
    description: str = (
        "SplitMultiValueProperty is a transformer that splits a "
        "multi-value property into multiple single-value properties."
    )
    _use_only_once: bool = True
    _need_changes = frozenset({})

    _object_property_template: str = """SELECT ?s ?o WHERE{{

                                ?s a <{subject_uri}> .
                                ?s <{property_uri}> ?o .
                                ?o a <{object_uri}> .

                            }}"""

    _datatype_property_template: str = """SELECT ?s ?o WHERE {{

                                ?s a <{subject_uri}> .
                                ?s <{property_uri}> ?o .
                                FILTER (datatype(?o) = <{object_uri}>)

                                }}"""

    _unknown_property_template: str = """SELECT ?s ?o WHERE {{

                                ?s a <{subject_uri}> .
                                ?s <{property_uri}> ?o .
                                FILTER NOT EXISTS {{ ?o a ?objectType }}
                                }}"""

    def transform(self, graph: Graph) -> None:
        # handle multi value type object properties
        for subject_uri, property_uri, value_types in Queries(graph).multi_value_type_property():
            for value_type_uri in value_types:
                _args = {
                    "subject_uri": subject_uri,
                    "property_uri": property_uri,
                    "object_uri": value_type_uri,
                }

                # Case 1: Unknown value type
                if value_type_uri == UNKNOWN_TYPE:
                    iterator = graph.query(self._unknown_property_template.format(**_args))

                # Case 2: Datatype value type
                elif value_type_uri.startswith(str(XSD)):
                    iterator = graph.query(self._datatype_property_template.format(**_args))

                # Case 3: Object value type
                else:
                    iterator = graph.query(self._object_property_template.format(**_args))

                for s, o in iterator:  # type: ignore [misc]
                    graph.remove((s, property_uri, o))
                    new_property = URIRef(f"{property_uri}_{remove_namespace_from_uri(value_type_uri)}")
                    graph.add((s, new_property, o))



from collections import Counter, defaultdict
from datetime import datetime
from pathlib import Path
from typing import cast

from rdflib import RDF
from rdflib import Literal as RdfLiteral

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._issues.warnings import PropertyValueTypeUndefinedWarning
from cognite.neat._rules.models._base_rules import MatchType
from cognite.neat._rules.models.data_types import AnyURI
from cognite.neat._rules.models.entities._single_value import UnknownEntity
from cognite.neat._rules.models.information import (
    InformationMetadata,
)
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.rdf_ import remove_namespace_from_uri

from ._base import DEFAULT_NON_EXISTING_NODE_TYPE, BaseRDFImporter

ORDERED_CLASSES_QUERY = """SELECT ?class (count(?s) as ?instances )
                           WHERE { ?s a ?class . }
                           group by ?class order by DESC(?instances)"""

INSTANCES_OF_CLASS_QUERY = """SELECT ?s WHERE { ?s a <class> . }"""

INSTANCE_PROPERTIES_DEFINITION = """SELECT ?property (count(?property) as ?occurrence) ?dataType ?objectType
                                    WHERE {<instance_id> ?property ?value .

                                           BIND(datatype(?value) AS ?dataType)

                                           OPTIONAL {?value rdf:type ?objectType .}}
                                    GROUP BY ?property ?dataType ?objectType"""


class InferenceImporter(BaseRDFImporter):
    """Infers rules from a triple store.

    Rules inference through analysis of knowledge graph provided in various formats.
    Use the factory methods to create a triple store from sources such as
    RDF files, JSON files, YAML files, XML files, or directly from a graph store.

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
        max_number_of_instance: Maximum number of instances to be used in inference
        prefix: Prefix to be used for the inferred model
    """

    @classmethod
    def from_graph_store(
        cls,
        store: NeatGraphStore,
        prefix: str = "inferred",
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
    ) -> "InferenceImporter":
        return super().from_graph_store(store, prefix, max_number_of_instance, non_existing_node_type)

    @classmethod
    def from_file(
        cls,
        filepath: Path,
        prefix: str = "inferred",
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
    ) -> "InferenceImporter":
        return super().from_file(filepath, prefix, max_number_of_instance, non_existing_node_type)

    @classmethod
    def from_json_file(
        cls,
        filepath: Path,
        prefix: str = "inferred",
        max_number_of_instance: int = -1,
    ) -> "InferenceImporter":
        raise NotImplementedError("JSON file format is not supported yet.")

    @classmethod
    def from_yaml_file(
        cls,
        filepath: Path,
        prefix: str = "inferred",
        max_number_of_instance: int = -1,
    ) -> "InferenceImporter":
        raise NotImplementedError("YAML file format is not supported yet.")

    @classmethod
    def from_xml_file(
        cls,
        filepath: Path,
        prefix: str = "inferred",
        max_number_of_instance: int = -1,
    ) -> "InferenceImporter":
        raise NotImplementedError("JSON file format is not supported yet.")

    def _to_rules_components(
        self,
    ) -> dict:
        """Convert RDF graph to dictionary defining data model and prefixes of the graph

        Args:
            graph: RDF graph to be converted to TransformationRules object
            max_number_of_instance: Max number of instances to be considered for each class

        Returns:
            Tuple of data model and prefixes of the graph
        """
        classes: dict[str, dict] = {}
        properties: dict[str, dict] = {}
        count_by_value_type_by_property: dict[str, dict[str, int]] = defaultdict(Counter)

        # Infers all the classes in the graph
        for class_uri, no_instances in self.graph.query(ORDERED_CLASSES_QUERY):  # type: ignore[misc]
            if (class_id := remove_namespace_from_uri(class_uri)) in classes:
                # handles cases when class id is already present in classes
                class_id = f"{class_id}_{len(classes)+1}"

            classes[class_id] = {
                "class_": class_id,
                "reference": class_uri,
                "match_type": MatchType.exact,
                "comment": f"Inferred from knowledge graph, where this class has <{no_instances}> instances",
            }

        # Infers all the properties of the class
        for class_id, class_definition in classes.items():
            for (instance,) in self.graph.query(  # type: ignore[misc]
                INSTANCES_OF_CLASS_QUERY.replace("class", class_definition["reference"])
                if self.max_number_of_instance < 0
                else INSTANCES_OF_CLASS_QUERY.replace("class", class_definition["reference"])
                + f" LIMIT {self.max_number_of_instance}"
            ):
                for property_uri, occurrence, data_type_uri, object_type_uri in self.graph.query(  # type: ignore[misc]
                    INSTANCE_PROPERTIES_DEFINITION.replace("instance_id", instance)
                ):  # type: ignore[misc]
                    # this is to skip rdf:type property
                    if property_uri == RDF.type:
                        continue

                    property_id = remove_namespace_from_uri(property_uri)

                    if value_type_uri := (data_type_uri or object_type_uri):
                        value_type_id = remove_namespace_from_uri(value_type_uri)

                    # this handles situations when property points to node that is not present in graph
                    else:
                        value_type_id = str(self.non_existing_node_type)

                        issue = PropertyValueTypeUndefinedWarning(
                            resource_type="Property",
                            identifier=f"{class_id}:{property_id}",
                            property_name=property_id,
                            default_action="Remove the property from the rules",
                            recommended_action="Make sure that graph is complete",
                        )

                        if issue not in self.issue_list:
                            self.issue_list.append(issue)

                    id_ = f"{class_id}:{property_id}"

                    definition = {
                        "class_": class_id,
                        "property_": property_id,
                        "max_count": cast(RdfLiteral, occurrence).value,
                        "value_type": value_type_id,
                        "reference": property_uri,
                    }

                    count_by_value_type_by_property[id_][value_type_id] += 1

                    # USE CASE 1: If property is not present in properties
                    if id_ not in properties:
                        properties[id_] = definition

                    # USE CASE 2: first time redefinition, value type change to multi
                    elif id_ in properties and definition["value_type"] not in properties[id_]["value_type"]:
                        properties[id_]["value_type"] = properties[id_]["value_type"] + " | " + definition["value_type"]

                    # USE CASE 3: existing but max count is different
                    elif (
                        id_ in properties
                        and definition["value_type"] in properties[id_]["value_type"]
                        and properties[id_]["max_count"] != definition["max_count"]
                    ):
                        properties[id_]["max_count"] = max(properties[id_]["max_count"], definition["max_count"])

        # Add comments
        for id_, property_ in properties.items():
            if id_ not in count_by_value_type_by_property:
                continue

            count_by_value_type = count_by_value_type_by_property[id_]
            count_list = sorted(count_by_value_type.items(), key=lambda item: item[1], reverse=True)
            # Make the comment more readable by adapting to the number of value types
            base_string = "<{value_type}> which occurs <{count}> times"
            if len(count_list) == 1:
                type_, count = count_list[0]
                counts_str = f"with value type {base_string.format(value_type=type_, count=count)} in the graph"
            elif len(count_list) == 2:
                first = base_string.format(value_type=count_list[0][0], count=count_list[0][1])
                second = base_string.format(value_type=count_list[1][0], count=count_list[1][1])
                counts_str = f"with value types {first} and {second} in the graph"
            else:
                first_part = ", ".join(
                    base_string.format(value_type=type_, count=count) for type_, count in count_list[:-1]
                )
                last = base_string.format(value_type=count_list[-1][0], count=count_list[-1][1])
                counts_str = f"with value types {first_part} and {last} in the graph"

            class_id = property_["class_"]
            property_id = property_["property_"]
            property_["comment"] = f"Class <{class_id}> has property <{property_id}> {counts_str}"

        return {
            "metadata": self._default_metadata().model_dump(),
            "classes": list(classes.values()),
            "properties": list(properties.values()),
        }

    def _default_metadata(self):
        return InformationMetadata(
            name="Inferred Model",
            creator="NEAT",
            version="inferred",
            created=datetime.now(),
            updated=datetime.now(),
            description="Inferred model from knowledge graph",
            prefix=self.prefix,
            namespace=DEFAULT_NAMESPACE,
        )



from typing import Literal, TypeAlias

MIMETypes: TypeAlias = Literal[
    "application/rdf+xml", "text/turtle", "application/n-triple", "application/n-quads", "application/trig"
]


def rdflib_to_mime_types(rdflib_format: str) -> str | None:
    """Convert an RDFlib format to a MIME type.

    Args:
        rdflib_format (str): The RDFlib format.

    Returns:
        MIMETypes: The MIME type.

    !!! note
        This will be replaced once new version of oxrdflib is released.

    """

    mapping = {
        "xml": "application/rdf+xml",
        "rdf": "application/rdf+xml",
        "owl": "application/rdf+xml",
        "n3": "application/n-triple",
        "ttl": "text/turtle",
        "turtle": "text/turtle",
        "nt": "application/n-triple",
        "nq": "application/n-quads",
        "nquads": "application/n-quads",
        "trig": "application/trig",
    }
    return mapping.get(rdflib_format, None)



import inspect
import sys
import warnings
from abc import ABC
from collections.abc import Collection, Hashable, Iterable, Iterator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, fields
from functools import total_ordering
from pathlib import Path
from types import UnionType
from typing import Any, ClassVar, Literal, TypeAlias, TypeVar, get_args, get_origin
from warnings import WarningMessage

import pandas as pd
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    DataModelId,
    PropertyId,
    ViewId,
)
from pydantic_core import ErrorDetails

from cognite.neat._utils.spreadsheet import SpreadsheetRead
from cognite.neat._utils.text import humanize_collection, to_camel, to_snake

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup
    from typing_extensions import Self
else:
    from typing import Self


__all__ = [
    "NeatIssue",
    "NeatError",
    "NeatWarning",
    "DefaultWarning",
    "NeatIssueList",
    "MultiValueError",
]

T_Identifier = TypeVar("T_Identifier", bound=Hashable)

T_ReferenceIdentifier = TypeVar("T_ReferenceIdentifier", bound=Hashable)

ResourceType: TypeAlias = (
    Literal[
        "view",
        "container",
        "view property",
        "container property",
        "space",
        "class",
        "property",
        "asset",
        "relationship",
        "data model",
        "edge",
        "node",
        "enum collection",
        "unknown",
    ]
    # String to handle all unknown types in different importers.
    | str
)


@total_ordering
@dataclass(unsafe_hash=True)
class NeatIssue:
    """This is the base class for all exceptions and warnings (issues) used in Neat."""

    extra: ClassVar[str | None] = None
    fix: ClassVar[str | None] = None

    def as_message(self, include_type: bool = True) -> str:
        """Return a human-readable message for the issue."""
        template = self.__doc__
        if not template:
            return "Missing"
        variables, has_all_optional = self._get_variables()

        msg = template.format(**variables)
        if self.extra and has_all_optional:
            msg += "\n" + self.extra.format(**variables)
        if self.fix:
            msg += f"\nFix: {self.fix.format(**variables)}"
        if include_type:
            name = type(self).__name__
            msg = f"{name}: {msg}"
        return msg

    def _get_variables(self) -> tuple[dict[str, str], bool]:
        variables: dict[str, str] = {}
        has_all_optional = True
        for name, var_ in vars(self).items():
            if var_ is None:
                has_all_optional = False
            elif isinstance(var_, str):
                variables[name] = var_
            elif isinstance(var_, Path):
                variables[name] = var_.as_posix()
            elif isinstance(var_, Collection):
                variables[name] = humanize_collection(var_)
            else:
                variables[name] = repr(var_)
        return variables, has_all_optional

    def dump(self) -> dict[str, Any]:
        """Return a dictionary representation of the issue."""
        variables = vars(self)
        output = {
            to_camel(key): self._dump_value(value)
            for key, value in variables.items()
            if not (value is None or key.startswith("_"))
        }
        output["NeatIssue"] = type(self).__name__
        return output

    @classmethod
    def _dump_value(cls, value: Any) -> list | int | bool | float | str | dict:
        from cognite.neat._rules.models.entities import Entity

        if isinstance(value, str | int | bool | float):
            return value
        elif isinstance(value, frozenset):
            return [cls._dump_value(item) for item in value]
        elif isinstance(value, Path):
            return value.as_posix()
        elif isinstance(value, tuple):
            return [cls._dump_value(item) for item in value]
        elif isinstance(value, ViewId | ContainerId):
            return value.dump(camel_case=True, include_type=True)
        elif isinstance(value, Entity):
            return value.dump()
        elif isinstance(value, PropertyId):
            return value.dump(camel_case=True)
        elif isinstance(value, DataModelId):
            return value.dump(camel_case=True, include_type=False)
        raise ValueError(f"Unsupported type: {type(value)}")

    @classmethod
    def load(cls, data: dict[str, Any]) -> "NeatIssue":
        """Create an instance of the issue from a dictionary."""
        from cognite.neat._issues.errors import _NEAT_ERRORS_BY_NAME, NeatValueError
        from cognite.neat._issues.warnings import _NEAT_WARNINGS_BY_NAME

        if "NeatIssue" not in data:
            raise NeatValueError("The data does not contain a NeatIssue key.")
        issue_type = data.pop("NeatIssue")
        args = {to_snake(key): value for key, value in data.items()}
        if issue_type in _NEAT_ERRORS_BY_NAME:
            return cls._load_values(_NEAT_ERRORS_BY_NAME[issue_type], args)
        elif issue_type in _NEAT_WARNINGS_BY_NAME:
            return cls._load_values(_NEAT_WARNINGS_BY_NAME[issue_type], args)
        else:
            raise NeatValueError(f"Unknown issue type: {issue_type}")

    @classmethod
    def _load_values(cls, neat_issue_cls: "type[NeatIssue]", data: dict[str, Any]) -> "NeatIssue":
        args: dict[str, Any] = {}
        for f in fields(neat_issue_cls):
            if f.name not in data:
                continue
            value = data[f.name]
            args[f.name] = cls._load_value(f.type, value)
        return neat_issue_cls(**args)

    @classmethod
    def _load_value(cls, type_: Any, value: Any) -> Any:
        from cognite.neat._rules.models.entities import Entity

        if isinstance(type_, UnionType) or get_origin(type_) is UnionType:
            args = get_args(type_)
            return cls._load_value(args[0], value)
        elif type_ is frozenset or get_origin(type_) is frozenset:
            subtype = get_args(type_)[0]
            return frozenset(cls._load_value(subtype, item) for item in value)
        elif type_ is Path:
            return Path(value)
        elif type_ is tuple or get_origin(type_) is tuple:
            subtype = get_args(type_)[0]
            return tuple(cls._load_value(subtype, item) for item in value)
        elif type_ is ViewId:
            return ViewId.load(value)
        elif type_ is DataModelId:
            return DataModelId.load(value)
        elif type_ is PropertyId:
            return PropertyId.load(value)
        elif type_ is ContainerId:
            return ContainerId.load(value)
        elif inspect.isclass(type_) and issubclass(type_, Entity):
            return type_.load(value)
        return value

    def __lt__(self, other: "NeatIssue") -> bool:
        if not isinstance(other, NeatIssue):
            return NotImplemented
        return (type(self).__name__, self.as_message()) < (type(other).__name__, other.as_message())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NeatIssue):
            return NotImplemented
        return (type(self).__name__, self.as_message()) == (type(other).__name__, other.as_message())


@dataclass(unsafe_hash=True)
class NeatError(NeatIssue, Exception):
    """This is the base class for all exceptions (errors) used in Neat."""

    @classmethod
    def from_pydantic_errors(cls, errors: list[ErrorDetails], **kwargs) -> "list[NeatError]":
        """Convert a list of pydantic errors to a list of Error instances.

        This is intended to be overridden in subclasses to handle specific error types.
        """
        all_errors: list[NeatError] = []
        read_info_by_sheet = kwargs.get("read_info_by_sheet")

        for error in errors:
            if error["type"] == "is_instance_of" and error["loc"][1] == "is-instance[SheetList]":
                # Skip the error for SheetList, as it is not relevant for the user. This is an
                # internal class used to have helper methods for a lists as .to_pandas()
                continue
            ctx = error.get("ctx")
            if isinstance(ctx, dict) and isinstance(multi_error := ctx.get("error"), MultiValueError):
                if read_info_by_sheet:
                    for caught_error in multi_error.errors:
                        cls._adjust_row_numbers(caught_error, read_info_by_sheet)  # type: ignore[arg-type]
                all_errors.extend(multi_error.errors)  # type: ignore[arg-type]
            elif isinstance(ctx, dict) and isinstance(single_error := ctx.get("error"), NeatError):
                if read_info_by_sheet:
                    cls._adjust_row_numbers(single_error, read_info_by_sheet)
                all_errors.append(single_error)
            elif len(error["loc"]) >= 4 and read_info_by_sheet:
                all_errors.append(RowError.from_pydantic_error(error, read_info_by_sheet))
            else:
                all_errors.append(DefaultPydanticError.from_pydantic_error(error))
        return all_errors

    @staticmethod
    def _adjust_row_numbers(caught_error: "NeatError", read_info_by_sheet: dict[str, SpreadsheetRead]) -> None:
        from cognite.neat._issues.errors._properties import PropertyDefinitionDuplicatedError
        from cognite.neat._issues.errors._resources import ResourceNotDefinedError

        reader = read_info_by_sheet.get("Properties", SpreadsheetRead())

        if isinstance(caught_error, PropertyDefinitionDuplicatedError) and caught_error.location_name == "rows":
            adjusted_row_number = (
                tuple(
                    reader.adjusted_row_number(row_no) if isinstance(row_no, int) else row_no
                    for row_no in caught_error.locations or []
                )
                or None
            )
            # The error is frozen, so we have to use __setattr__ to change the row number
            object.__setattr__(caught_error, "locations", adjusted_row_number)
        elif isinstance(caught_error, RowError):
            # Adjusting the row number to the actual row number in the spreadsheet
            new_row = reader.adjusted_row_number(caught_error.row)
            # The error is frozen, so we have to use __setattr__ to change the row number
            object.__setattr__(caught_error, "row", new_row)
        elif isinstance(caught_error, ResourceNotDefinedError):
            if isinstance(caught_error.row_number, int) and caught_error.sheet_name == "Properties":
                new_row = reader.adjusted_row_number(caught_error.row_number)
                object.__setattr__(caught_error, "row_number", new_row)


@dataclass(unsafe_hash=True)
class DefaultPydanticError(NeatError, ValueError):
    """{type}: {msg} [loc={loc}]"""

    type: str
    loc: tuple[int | str, ...]
    msg: str

    @classmethod
    def from_pydantic_error(cls, error: ErrorDetails) -> "DefaultPydanticError":
        return cls(
            type=error["type"],
            loc=error["loc"],
            msg=error["msg"],
        )

    def as_message(self, include_type: bool = True) -> str:
        if self.loc and len(self.loc) == 1:
            return f"{self.loc[0]} sheet: {self.msg}"
        elif self.loc and len(self.loc) == 2:
            return f"{self.loc[0]} sheet field/column <{self.loc[1]}>: {self.msg}"
        else:
            return self.msg


@dataclass(unsafe_hash=True)
class RowError(NeatError, ValueError):
    """In {sheet_name}, row={row}, column={column}: {msg}. [type={type}, input_value={input}]"""

    extra = "For further information visit {url}"

    sheet_name: str
    column: str
    row: int
    type: str
    msg: str
    input: Any
    url: str | None = None

    @classmethod
    def from_pydantic_error(
        cls,
        error: ErrorDetails,
        read_info_by_sheet: dict[str, SpreadsheetRead] | None = None,
    ) -> Self:
        sheet_name, _, row, column, *__ = error["loc"]
        reader = (read_info_by_sheet or {}).get(str(sheet_name), SpreadsheetRead())
        return cls(
            sheet_name=str(sheet_name),
            column=str(column),
            row=reader.adjusted_row_number(int(row)),
            type=error["type"],
            msg=error["msg"],
            input=error.get("input"),
            url=str(url) if (url := error.get("url")) else None,
        )

    def as_message(self, include_type: bool = True) -> str:
        input_str = str(self.input) if self.input is not None else ""
        input_str = input_str[:50] + "..." if len(input_str) > 50 else input_str
        output = (
            f"In {self.sheet_name}, row={self.row}, column={self.column}: {self.msg}. "
            f"[type={self.type}, input_value={input_str}]"
        )
        if self.url:
            output += f" For further information visit {self.url}"
        return output


@dataclass(unsafe_hash=True)
class NeatWarning(NeatIssue, UserWarning):
    """This is the base class for all warnings used in Neat."""

    @classmethod
    def from_warning(cls, warning: WarningMessage) -> "NeatWarning":
        """Create a NeatWarning from a WarningMessage."""
        return DefaultWarning.from_warning_message(warning)


@dataclass(unsafe_hash=True)
class DefaultWarning(NeatWarning):
    """{category}: {warning}"""

    extra = "Source: {source}"

    warning: str
    category: str
    source: str | None = None

    @classmethod
    def from_warning_message(cls, warning: WarningMessage) -> NeatWarning:
        if isinstance(warning.message, NeatWarning):
            return warning.message

        return cls(
            warning=str(warning.message),
            category=warning.category.__name__,
            source=warning.source,
        )

    def as_message(self, include_type: bool = True) -> str:
        return str(self.warning)


T_NeatIssue = TypeVar("T_NeatIssue", bound=NeatIssue)


class NeatIssueList(list, Sequence[T_NeatIssue], ABC):
    """This is a generic list of NeatIssues."""

    def __init__(
        self, issues: Sequence[T_NeatIssue] | None = None, title: str | None = None, action: str | None = None
    ):
        super().__init__(issues or [])
        self.title = title
        self.action = action

    @property
    def errors(self) -> Self:
        """Return all the errors in this list."""
        return type(self)([issue for issue in self if isinstance(issue, NeatError)])  # type: ignore[misc]

    @property
    def has_errors(self) -> bool:
        """Return True if this list contains any errors."""
        return any(isinstance(issue, NeatError) for issue in self)

    @property
    def warnings(self) -> Self:
        """Return all the warnings in this list."""
        return type(self)([issue for issue in self if isinstance(issue, NeatWarning)])  # type: ignore[misc]

    def as_errors(self, operation: str = "Operation failed") -> ExceptionGroup:
        """Return an ExceptionGroup with all the errors in this list."""
        return ExceptionGroup(
            operation,
            [issue for issue in self if isinstance(issue, NeatError)],
        )

    def trigger_warnings(self) -> None:
        """Trigger all warnings in this list."""
        for warning in [issue for issue in self if isinstance(issue, NeatWarning)]:
            warnings.warn(warning, stacklevel=2)

    def to_pandas(self) -> pd.DataFrame:
        """Return a pandas DataFrame representation of this list."""
        return pd.DataFrame([issue.dump() for issue in self])

    def _repr_html_(self) -> str | None:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]

    def as_exception(self) -> "MultiValueError":
        """Return a MultiValueError with all the errors in this list."""
        return MultiValueError(self.errors)


class MultiValueError(ValueError):
    """This is a container for multiple errors.

    It is used in the pydantic field_validator/model_validator to collect multiple errors, which
    can then be caught in a try-except block and returned as an IssueList.

    """

    def __init__(self, errors: Sequence[NeatIssue]):
        self.errors = list(errors)


class IssueList(NeatIssueList[NeatIssue]):
    """This is a list of NeatIssues."""

    def _repr_html_(self) -> str | None:
        if not self:
            return "<p>'No issues found'</p>"
        df = self.to_pandas()
        agg_df = df["NeatIssue"].value_counts().to_frame()
        if len(agg_df) < 10:
            return agg_df._repr_html_()  # type: ignore[operator]
        else:
            return agg_df.head()._repr_html_()  # type: ignore[operator]


T_Cls = TypeVar("T_Cls")


def _get_subclasses(cls_: type[T_Cls], include_base: bool = False) -> Iterable[type[T_Cls]]:
    """Get all subclasses of a class."""
    if include_base:
        yield cls_
    for s in cls_.__subclasses__():
        yield s
        yield from _get_subclasses(s, False)


@contextmanager
def catch_warnings(
    issues: IssueList | None = None,
    warning_cls: type[NeatWarning] = DefaultWarning,
) -> Iterator[None]:
    """Catch warnings and append them to the issues list."""
    with warnings.catch_warnings(record=True) as warning_logger:
        warnings.simplefilter("always")
        try:
            yield None
        finally:
            if warning_logger and issues is not None:
                issues.extend([warning_cls.from_warning(warning) for warning in warning_logger])  # type: ignore[misc]



from collections import Counter, defaultdict
from collections.abc import Iterator, MutableSequence, Sequence
from pathlib import Path
from typing import Any, Generic, SupportsIndex, TypeVar, cast, get_args, overload

import pandas as pd
from pydantic import BaseModel, GetCoreSchemaHandler, field_validator
from pydantic_core import core_schema
from pydantic_core.core_schema import ValidationInfo

from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules.models._base_rules import ClassRef, PropertyRef
from cognite.neat._rules.models.entities import ClassEntity, Undefined

T_Mapping = TypeVar("T_Mapping", bound=ClassRef | PropertyRef)


class Mapping(BaseModel, Generic[T_Mapping]):
    source: T_Mapping
    destination: T_Mapping


class MappingList(list, MutableSequence[Mapping[T_Mapping]]):
    @classmethod
    def __get_pydantic_core_schema__(cls, source: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        if args := get_args(source):
            item_type = args[0]
        else:
            # Someone use SheetList without specifying the type
            raise TypeError("SheetList must be used with a type argument, e.g., SheetList[InformationProperty]")

        instance_schema = core_schema.is_instance_schema(cls)
        sequence_row_schema = handler.generate_schema(Sequence[item_type])  # type: ignore[valid-type]

        non_instance_schema = core_schema.no_info_after_validator_function(MappingList, sequence_row_schema)
        return core_schema.union_schema([instance_schema, non_instance_schema])

    def as_destination_by_source(self) -> dict[T_Mapping, T_Mapping]:
        return {mapping.source: mapping.destination for mapping in self}

    def to_pandas(self, drop_na_columns: bool = True, include: list[str] | None = None) -> pd.DataFrame:
        """Converts ResourceDict to pandas DataFrame."""
        df = pd.DataFrame([entity.model_dump() for entity in self])
        if drop_na_columns:
            df = df.dropna(axis=1, how="all")
        if include is not None:
            df = df[include]
        return df

    def _repr_html_(self) -> str:
        """Returns HTML representation of ResourceDict."""
        return self.to_pandas(drop_na_columns=True)._repr_html_()  # type: ignore[operator]

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[Mapping[T_Mapping]]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> Mapping[T_Mapping]: ...

    @overload
    def __getitem__(self, index: slice) -> "MappingList[T_Mapping]": ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "Mapping[T_Mapping] | MappingList[T_Mapping]":
        if isinstance(index, slice):
            return MappingList[T_Mapping](super().__getitem__(index))
        return super().__getitem__(index)


class RuleMapping(BaseModel):
    properties: MappingList[PropertyRef]
    classes: MappingList[ClassRef]

    @field_validator("properties", "classes", mode="before")
    def as_mapping_list(cls, value: Sequence[Any], info: ValidationInfo) -> Any:
        if isinstance(value, Sequence) and not isinstance(value, MappingList):
            annotation = cast(type, cls.model_fields[info.field_name].annotation)  # type: ignore[index]
            ref_cls = get_args(annotation)[0]
            return annotation([Mapping[ref_cls].model_validate(item) for item in value])  # type: ignore[valid-type, index]
        return value

    @classmethod
    def load_spreadsheet(
        cls, path: str | Path, source_prefix: str | None = None, destination_prefix: str | None = None
    ) -> "RuleMapping":
        """Loads mapping from Excel spreadsheet.

        This method expects four columns in the spreadsheet. The first two columns are the source class and
        property, and the last two columns are the destination class and property. The method will create
        a mapping for each row in the spreadsheet.

        The class mapping will be inferred from the property mappings. If a source class has multiple
        destination classes, the most common destination class will be used.

        Args:
            path: Path to Excel spreadsheet.
            source_prefix: Default prefix for source classes.
            destination_prefix: Default prefix for destination classes.

        Returns:
            Mapping object.

        """
        df = pd.read_excel(path).dropna(axis=1, how="all")
        properties = MappingList[PropertyRef]()
        destination_classes_by_source: dict[ClassEntity, Counter[ClassEntity]] = defaultdict(Counter)
        for _, row in df.iterrows():
            if len(row) < 4:
                raise NeatValueError(f"Row {row} is not valid. Expected 4 columns, got {len(row)}")

            if any(pd.isna(row.iloc[:4])):
                continue
            source_class, source_property, destination_class, destination_property = row.iloc[:4]
            source_entity = ClassEntity.load(source_class, prefix=source_prefix or Undefined)
            destination_entity = ClassEntity.load(destination_class, prefix=destination_prefix or Undefined)
            properties.append(
                Mapping(
                    source=PropertyRef(Class=source_entity, Property=source_property),
                    destination=PropertyRef(Class=destination_entity, Property=destination_property),
                )
            )
            destination_classes_by_source[source_entity][destination_entity] += 1

        classes = MappingList[ClassRef]()
        for source_entity, destination_classes in destination_classes_by_source.items():
            destination_entity = destination_classes.most_common(1)[0][0]
            classes.append(
                Mapping(source=ClassRef(Class=source_entity), destination=ClassRef(Class=destination_entity))
            )

        return cls(properties=properties, classes=classes)



from collections.abc import Callable, Set
from datetime import datetime, timezone
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import Sequence, SequenceFilter, SequenceList
from rdflib import RDF, Literal, Namespace

from cognite.neat._graph.models import Triple

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix


class SequencesExtractor(ClassicCDFBaseExtractor[Sequence]):
    """Extract data from Cognite Data Fusions Sequences into Neat.

    Args:
        items (Iterable[Sequence]): An iterable of items.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        to_type (Callable[[Sequence], str | None], optional): A function to convert an item to a type.
            Defaults to None. If None or if the function returns None, the asset will be set to the default type.
        total (int, optional): The total number of items to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of items to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): If you are unpacking metadata, then
           values in this set will be skipped.
    """

    _default_rdf_type = "Sequence"

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Sequence], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        total = client.sequences.aggregate_count(
            filter=SequenceFilter(data_set_ids=[{"externalId": data_set_external_id}])
        )
        return cls(
            client.sequences(data_set_external_ids=data_set_external_id),
            total=total,
            namespace=namespace,
            to_type=to_type,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    @classmethod
    def from_hierarchy(
        cls,
        client: CogniteClient,
        root_asset_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Sequence], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        total = client.sequences.aggregate_count(
            filter=SequenceFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )

        return cls(
            client.sequences(asset_subtree_external_ids=[root_asset_external_id]),
            namespace,
            to_type,
            total,
            limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Sequence], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        sequences = SequenceList.load(Path(file_path).read_text())
        return cls(
            sequences,
            total=len(sequences),
            namespace=namespace,
            to_type=to_type,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    def _item2triples(self, sequence: Sequence) -> list[Triple]:
        id_ = self.namespace[f"{InstanceIdPrefix.sequence}{sequence.id}"]

        type_ = self._get_rdf_type(sequence)
        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, self.namespace[type_])]

        # Create attributes

        if sequence.external_id:
            triples.append((id_, self.namespace.external_id, Literal(sequence.external_id)))

        if sequence.name:
            triples.append((id_, self.namespace.name, Literal(sequence.name)))

        if sequence.metadata:
            triples.extend(self._metadata_to_triples(id_, sequence.metadata))

        if sequence.description:
            triples.append((id_, self.namespace.description, Literal(sequence.description)))

        if sequence.created_time:
            triples.append(
                (
                    id_,
                    self.namespace.created_time,
                    Literal(datetime.fromtimestamp(sequence.created_time / 1000, timezone.utc)),
                )
            )

        if sequence.last_updated_time:
            triples.append(
                (
                    id_,
                    self.namespace.last_updated_time,
                    Literal(datetime.fromtimestamp(sequence.last_updated_time / 1000, timezone.utc)),
                )
            )

        if sequence.data_set_id:
            triples.append(
                (
                    id_,
                    self.namespace.data_set_id,
                    self.namespace[f"{InstanceIdPrefix.data_set}{sequence.data_set_id}"],
                )
            )

        if sequence.asset_id:
            triples.append(
                (
                    id_,
                    self.namespace.asset,
                    self.namespace[f"{InstanceIdPrefix.asset}{sequence.asset_id}"],
                )
            )

        return triples



from collections.abc import Collection
from typing import Literal, cast

from cognite.client.data_classes.data_modeling import DataModelIdentifier

from cognite.neat._issues._base import IssueList
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.models.information._rules_input import InformationInputRules
from cognite.neat._rules.transformers import ReduceCogniteModel, ToCompliantEntities, ToExtension

from ._state import SessionState
from .exceptions import intercept_session_exceptions


@intercept_session_exceptions
class PrepareAPI:
    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.data_model = DataModelPrepareAPI(state, verbose)


@intercept_session_exceptions
class DataModelPrepareAPI:
    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def cdf_compliant_external_ids(self) -> None:
        """Convert data model component external ids to CDF compliant entities."""
        if input := self._state.information_input_rule:
            output = ToCompliantEntities().transform(input)
            self._state.input_rules.append(
                ReadRules(
                    rules=cast(InformationInputRules, output.get_rules()),
                    issues=IssueList(),
                    read_context={},
                )
            )

    def to_enterprise(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "My",
        dummy_property: str = "GUID",
    ) -> None:
        """Uses the current data model as a basis to create enterprise data model

        Args:
            data_model_id: The enterprise data model id that is being created
            org_name: Organization name to use for the views in the enterprise data model.
            dummy_property: The dummy property to use as placeholder for the views in the new data model.

        !!! note "Enterprise Data Model Creation"
            Always create an enterprise data model from a Cognite Data Model as this will
            assure all the Cognite Data Fusion applications to run smoothly, such as
            - Search
            - Atlas AI
            - ...

        """
        if dms := self._state.last_verified_dms_rules:
            output = ToExtension(
                new_model_id=data_model_id,
                org_name=org_name,
                type_="enterprise",
                dummy_property=dummy_property,
            ).transform(dms)
            self._state.verified_rules.append(output.rules)

    def to_solution(
        self,
        data_model_id: DataModelIdentifier,
        org_name: str = "My",
        mode: Literal["read", "write"] = "read",
        dummy_property: str = "dummy",
    ) -> None:
        """Uses the current data model as a basis to create solution data model

        Args:
            data_model_id: The solution data model id that is being created.
            org_name: Organization name to use for the views in the new data model.
            mode: The mode of the solution data model. Can be either "read" or "write".
            dummy_property: The dummy property to use as placeholder for the views in the new data model.

        !!! note "Solution Data Model Mode"
            The read-only solution model will only be able to read from the existing containers
            from the enterprise data model, therefore the solution data model will not have
            containers in the solution data model space. Meaning the solution data model views
            will be read-only.

            The write mode will have additional containers in the solution data model space,
            allowing in addition to reading through the solution model views, also writing to
            the containers in the solution data model space.

        """
        if dms := self._state.last_verified_dms_rules:
            output = ToExtension(
                new_model_id=data_model_id,
                org_name=org_name,
                type_="solution",
                mode=mode,
                dummy_property=dummy_property,
            ).transform(dms)
            self._state.verified_rules.append(output.rules)

    def reduce(self, drop: Collection[Literal["3D", "Annotation", "BaseViews"] | str]) -> None:
        """This is a special method that allow you to drop parts of the data model.
        This only applies to Cognite Data Models.

        Args:
            drop: What to drop from the data model. The values 3D, Annotation, and BaseViews are special values that
                drops multiple views at once. You can also pass externalIds of views to drop individual views.

        """
        if dms := self._state.last_verified_dms_rules:
            output = ReduceCogniteModel(drop).transform(dms)
            self._state.verified_rules.append(output.rules)



import warnings
from collections import defaultdict
from collections.abc import Iterable
from typing import Literal, cast, overload

from rdflib import RDF, Graph, Namespace, URIRef
from rdflib import Literal as RdfLiteral
from rdflib.query import ResultRow

from cognite.neat._constants import UNKNOWN_TYPE
from cognite.neat._graph.models import InstanceType
from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules.models.entities import ClassEntity
from cognite.neat._rules.models.information import InformationRules
from cognite.neat._utils.rdf_ import remove_namespace_from_uri

from ._construct import build_construct_query


class Queries:
    """Helper class for storing standard queries for the graph store."""

    def __init__(self, graph: Graph, rules: InformationRules | None = None):
        self.graph = graph
        self.rules = rules

    def summarize_instances(self) -> list[tuple]:
        """Summarize instances in the graph store by class and count"""

        query_statement = """ SELECT ?class (COUNT(?instance) AS ?instanceCount)
                             WHERE {
                             ?instance a ?class .
                             }
                             GROUP BY ?class
                             ORDER BY DESC(?instanceCount) """

        return [
            (
                remove_namespace_from_uri(cast(URIRef, cast(tuple, res)[0])),
                cast(RdfLiteral, cast(tuple, res)[1]).value,
            )
            for res in list(self.graph.query(query_statement))
        ]

    def list_instances_ids_of_class(self, class_uri: URIRef, limit: int = -1) -> list[URIRef]:
        """Get instances ids for a given class

        Args:
            class_uri: Class for which instances are to be found
            limit: Max number of instances to return, by default -1 meaning all instances

        Returns:
            List of class instance URIs
        """
        query_statement = "SELECT DISTINCT ?subject WHERE { ?subject a <class> .} LIMIT X".replace(
            "class", class_uri
        ).replace("LIMIT X", "" if limit == -1 else f"LIMIT {limit}")
        return [cast(tuple, res)[0] for res in list(self.graph.query(query_statement))]

    def list_instances_of_type(self, class_uri: URIRef) -> list[ResultRow]:
        """Get all triples for instances of a given class

        Args:
            class_uri: Class for which instances are to be found

        Returns:
            List of triples for instances of the given class
        """
        query = (
            f"SELECT ?instance ?prop ?value "
            f"WHERE {{ ?instance rdf:type <{class_uri}> . ?instance ?prop ?value . }} order by ?instance "
        )

        # Select queries gives an iterable of result rows
        return cast(list[ResultRow], list(self.graph.query(query)))

    def triples_of_type_instances(self, rdf_type: str | URIRef) -> list[tuple[str, str, str]]:
        """Get all triples of a given type.

        This method assumes the graph has been transformed into the default namespace.
        """
        if isinstance(rdf_type, URIRef):
            rdf_uri = rdf_type
        elif isinstance(rdf_type, str) and self.rules:
            rdf_uri = self.rules.metadata.namespace[rdf_type]
        else:
            warnings.warn(
                "Unknown namespace. Please either provide a URIRef or set the rules of the store.",
                stacklevel=2,
            )
            return []

        query = (
            "SELECT ?instance ?prop ?value "
            f"WHERE {{ ?instance a <{rdf_uri}> . ?instance ?prop ?value . }} "
            "order by ?instance"
        )

        result = self.graph.query(query)

        # We cannot include the RDF.type in case there is a neat:type property
        return [remove_namespace_from_uri(list(triple)) for triple in result if triple[1] != RDF.type]  # type: ignore[misc, index, arg-type]

    def types_with_property(self, property_uri: URIRef) -> list[URIRef]:
        """Check if a property exists in the graph store

        Args:
            property_uri: Property URI to check

        Returns:
            True if property exists, False otherwise
        """
        query = f"SELECT DISTINCT ?t WHERE {{ ?s <{property_uri}> ?o ; a ?t}} Limit 1"
        return cast(list[URIRef], [t[0] for t in self.graph.query(query)])  # type: ignore[index]

    def has_namespace(self, namespace: Namespace) -> bool:
        """Check if a namespace exists in the graph store

        Args:
            namespace: Namespace to check

        Returns:
            True if namespace exists, False otherwise
        """
        query = f"ASK WHERE {{ ?s ?p ?o . FILTER(STRSTARTS(STR(?p), STR(<{namespace}>))) }}"
        return bool(self.graph.query(query))

    def has_type(self, type_: URIRef) -> bool:
        """Check if a type exists in the graph store

        Args:
            type_: Type to check

        Returns:
            True if type exists, False otherwise
        """
        query = f"ASK WHERE {{ ?s a <{type_}> }}"
        return bool(self.graph.query(query))

    def describe(
        self,
        instance_id: URIRef,
        instance_type: str | None = None,
        property_renaming_config: dict | None = None,
        property_types: dict[str, EntityTypes] | None = None,
    ) -> tuple[str, dict[str | InstanceType, list[str]]] | None:
        """DESCRIBE instance for a given class from the graph store

        Args:
            instance_id: Instance id for which we want to generate query
            instance_type: Type of the instance, default None (will be inferred from triples)
            property_renaming_config: Dictionary to rename properties, default None (no renaming)
            property_types: Dictionary of property types, default None (helper for removal of namespace)


        Returns:
            Dictionary of instance properties
        """
        property_values: dict[str, list[str]] = defaultdict(list)
        identifier = remove_namespace_from_uri(instance_id, validation="prefix")
        for _, predicate, object_ in cast(list[ResultRow], self.graph.query(f"DESCRIBE <{instance_id}>")):
            if object_.lower() in [
                "",
                "none",
                "nan",
                "null",
            ]:
                continue

            # set property
            if property_renaming_config and predicate != RDF.type:
                property_ = property_renaming_config.get(
                    predicate, remove_namespace_from_uri(predicate, validation="prefix")
                )
            elif not property_renaming_config and predicate != RDF.type:
                property_ = remove_namespace_from_uri(predicate, validation="prefix")
            else:
                property_ = RDF.type

            # set value
            # if it is URIRef and property type is object property, we need to remove namespace
            # if it URIref but we are doing this into data type property, we do not remove namespace
            # case 1 for RDF type we remove namespace
            if property_ == RDF.type:
                value = remove_namespace_from_uri(object_, validation="prefix")

            # case 2 for define object properties we remove namespace
            elif (
                isinstance(object_, URIRef)
                and property_types
                and property_types.get(property_, None) == EntityTypes.object_property
            ):
                value = remove_namespace_from_uri(object_, validation="prefix")

            # case 3 when property type is not defined and returned value is URIRef we remove namespace
            elif isinstance(object_, URIRef) and not property_types:
                value = remove_namespace_from_uri(object_, validation="prefix")

            # case 4 for data type properties we do not remove namespace but keep the entire value
            # but we drop the datatype part, and keep everything to be string (data loader will do the conversion)
            # for value type it expects (if possible)
            else:
                value = str(object_)

            # add type to the dictionary
            if predicate != RDF.type:
                property_values[property_].append(value)
            else:
                # guarding against multiple rdf:type values as this is not allowed in CDF
                if RDF.type not in property_values:
                    property_values[RDF.type].append(instance_type if instance_type else value)
                else:
                    # we should not have multiple rdf:type values
                    continue

        if property_values:
            return (
                identifier,
                property_values,
            )
        else:
            return None

    def construct_instances_of_class(
        self,
        class_: str,
        properties_optional: bool = True,
        instance_id: URIRef | None = None,
    ) -> list[tuple[str, str, str]]:
        """CONSTRUCT instances for a given class from the graph store

        Args:
            class_: Class entity for which we want to generate query
            properties_optional: Whether to make all properties optional, default True
            instance_ids: List of instance ids to filter on, default None (all)

        Returns:
            List of triples for instances of the given class
        """

        if self.rules and (
            query := build_construct_query(
                class_=ClassEntity(prefix=self.rules.metadata.prefix, suffix=class_),
                graph=self.graph,
                rules=self.rules,
                properties_optional=properties_optional,
                instance_id=instance_id,
            )
        ):
            result = self.graph.query(query)

            # We cannot include the RDF.type in case there is a neat:type property
            return [remove_namespace_from_uri(cast(ResultRow, triple)) for triple in result if triple[1] != RDF.type]  # type: ignore[misc, index]
        else:
            warnings.warn(
                "No rules found for the graph store, returning empty list.",
                stacklevel=2,
            )
            return []

    def list_triples(self, limit: int = 25) -> list[ResultRow]:
        """List triples in the graph store

        Args:
            limit: Max number of triples to return, by default 25

        Returns:
            List of triples
        """
        query = f"SELECT ?subject ?predicate ?object WHERE {{ ?subject ?predicate ?object }} LIMIT {limit}"
        return cast(list[ResultRow], list(self.graph.query(query)))

    @overload
    def list_types(self, remove_namespace: Literal[False] = False, limit: int = 25) -> list[ResultRow]: ...

    @overload
    def list_types(self, remove_namespace: Literal[True], limit: int = 25) -> list[str]: ...

    def list_types(self, remove_namespace: bool = False, limit: int = 25) -> list[ResultRow] | list[str]:
        """List types in the graph store

        Args:
            limit: Max number of types to return, by default 25
            remove_namespace: Whether to remove the namespace from the type, by default False

        Returns:
            List of types
        """
        query = f"SELECT DISTINCT ?type WHERE {{ ?subject a ?type }} LIMIT {limit}"
        result = cast(list[ResultRow], list(self.graph.query(query)))
        if remove_namespace:
            return [remove_namespace_from_uri(res[0]) for res in result]
        return result

    def multi_value_type_property(
        self,
    ) -> Iterable[tuple[URIRef, URIRef, list[URIRef]]]:
        query = """SELECT ?sourceType ?property
                          (GROUP_CONCAT(DISTINCT STR(?valueType); SEPARATOR=",") AS ?valueTypes)

                   WHERE {{
                       ?s ?property ?o .
                       ?s a ?sourceType .
                       OPTIONAL {{ ?o a ?type }}

                       # Key part to determine value type: either object, data or unknown
                       BIND(   IF(isLiteral(?o),DATATYPE(?o),
                               IF(BOUND(?type), ?type,
                                               <{unknownType}>)) AS ?valueType)
                   }}

                   GROUP BY ?sourceType ?property
                   HAVING (COUNT(DISTINCT ?valueType) > 1)"""

        for (
            source_type,
            property_,
            value_types,
        ) in cast(
            ResultRow,
            self.graph.query(query.format(unknownType=str(UNKNOWN_TYPE))),
        ):
            yield cast(URIRef, source_type), cast(URIRef, property_), [URIRef(uri) for uri in value_types.split(",")]



from typing import cast

from rdflib import Graph

from cognite.neat._rules.importers._rdf._shared import (
    clean_up_classes,
    make_classes_compliant,
    parse_raw_classes_dataframe,
)


def parse_owl_classes(graph: Graph, language: str = "en") -> list[dict]:
    """Parse owl classes from graph to pandas dataframe.

    Args:
        graph: Graph containing owl classes
        language: Language to use for parsing, by default "en"

    Returns:
        Dataframe containing owl classes

    !!! note "Compliant OWL classes"
        This makes the method very opinionated, but results in a compliant classes.
    """

    query = """
        SELECT ?class ?name ?description ?parentClass ?reference ?match ?comment
        WHERE {
        ?class a owl:Class .
        OPTIONAL {?class rdfs:subClassOf ?parentClass }.
        OPTIONAL {?class rdfs:label ?name }.
        OPTIONAL {?class rdfs:comment ?description} .
        FILTER (!isBlank(?class))
        FILTER (!bound(?parentClass) || !isBlank(?parentClass))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        BIND(?class AS ?reference)
    }
    """

    # create raw dataframe

    raw_df = parse_raw_classes_dataframe(cast(list[tuple], list(graph.query(query.replace("en", language)))))
    if raw_df.empty:
        return []

    # group values and clean up
    processed_df = clean_up_classes(raw_df)

    # make compliant
    processed_df = make_classes_compliant(processed_df, importer="OWL")

    # Make Parent Class list elements into string joined with comma
    processed_df["Parent Class"] = processed_df["Parent Class"].apply(
        lambda x: ", ".join(x) if isinstance(x, list) and x else None
    )

    return processed_df.dropna(axis=0, how="all").replace(float("nan"), None).to_dict(orient="records")



from rdflib import Graph, Namespace, URIRef
from rdflib.query import ResultRow
from rdflib.term import Identifier

from ._base import BaseTransformer


# TODO: Handle the cse when value is None, which will not make the TextObject resolve
class TwoHopFlattener(BaseTransformer):
    """
    Transformer that will flatten the distance between a source node, an intermediate connecting node, and a
    target property that is connected to the intermediate node.
    The transformation result is that the target property is attached directly to the source node, instead of having
    to go via the intermediate node.
    The user can also provide a flag to decide if the intermediate node should be removed from the graph or not
    after connecting the target property to the source node.

        Ex. TwoHopFlattener:

        Graph before flattening (with deletion of intermediate node):
        node(A, rdf:type(Pump)) -(predicate("vendor"))>
                                node(B, rdf:type(TextObject)) -(predicate("value"))> Literal("CompanyX")

        Graph after flattening nodes with destination_node_type = rdf:type(TextObject), property_predicate = :value,
        and property_name = "value":

        node(A, rdf:type(Pump)) -(predicate("vendor"))> Literal("CompanyX")

    Args:
        destination_node_type: RDF.type of edge Node
        property_predicate: Predicate to use when resolving the value from the edge node
        property_name: name of the property that the intermediate node is pointing to
        delete_connecting_node: bool if the intermediate Node and Edge between source Node
                                and target property should be deleted. Defaults to True.
    """

    description: str = "Prunes the graph of specified node types that do not have connections to other nodes."
    _query_template: str = """SELECT ?sourceNode ?property ?destinationNode ?value WHERE {{
                                     ?sourceNode ?property ?destinationNode .
                                     ?destinationNode a <{destination_node_type}> .
                                     ?destinationNode <{property_predicate}> ?{property_name} . }}"""

    def __init__(
        self,
        destination_node_type: URIRef,
        property_predicate: Namespace,
        property_name: str,
        delete_connecting_node: bool = True,
    ):
        self.destination_node_type = destination_node_type
        self.property_predicate = property_predicate
        self.property_name = property_name
        self.delete_connecting_node = delete_connecting_node

    def transform(self, graph: Graph) -> None:
        nodes_to_delete: list[Identifier] = []

        graph_traversals = list(
            graph.query(
                self._query_template.format(
                    destination_node_type=self.destination_node_type,
                    property_predicate=self.property_predicate,
                    property_name=self.property_name,
                )
            )
        )

        for path in graph_traversals:
            if isinstance(path, ResultRow):
                source_node, predicate, destination_node, property_value = path.asdict().values()

                # Create new connection from source node to value
                graph.add((source_node, predicate, property_value))
                nodes_to_delete.append(destination_node)

            if self.delete_connecting_node:
                for node in nodes_to_delete:
                    # Remove edge triples to node
                    graph.remove((None, None, node))
                    # Remove node triple
                    graph.remove((node, None, None))


class PruneDanglingNodes(BaseTransformer):
    """
    Knowledge graph pruner and resolver. Will remove rdf triples from graph that does not have connections
    to other nodes, and traverse graph for specified types to resolve the value in the final node and link it to
    the source node.

        Ex. PruneDanglingNodes

        Graph before pruning:
        node(A, rdf:type(Pump)) -> node(B, rdf:type(Disc)),
        node(C, rdf:type(Disc))

        Graph after pruning of nodes rdf:type(Disc):

        node(A, rd:type(Pump)) -> node(B, rdf:type(Disc))

    Args:
        node_prune_types: list of RDF types to prune from the Graph if they are stand-alone Nodes
    """

    description: str = "Prunes the graph of specified rdf types that do not have connections to other nodes."
    _query_template = """
                    SELECT ?subject
                    WHERE {{
                        ?subject a <{rdf_type}> .
                        FILTER NOT EXISTS {{ ?s ?p ?subject }}
                    }}
            """

    def __init__(
        self,
        node_prune_types: list[URIRef],
    ):
        self.node_prune_types = node_prune_types

    def transform(self, graph: Graph) -> None:
        for object_type in self.node_prune_types:
            nodes_without_neighbours = list(graph.query(self._query_template.format(rdf_type=object_type)))

            for node in nodes_without_neighbours:
                # Remove node and its property triples in the graph
                if isinstance(node, ResultRow):
                    graph.remove(triple=(node["subject"], None, None))



import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager, suppress
from datetime import datetime
from typing import Any, Generic, Literal

from pydantic import ValidationError
from rdflib import Namespace

from cognite.neat._issues import IssueList, NeatError, NeatWarning
from cognite.neat._rules._shared import ReadRules, T_InputRules
from cognite.neat._utils.auxiliary import class_html_doc


class BaseImporter(ABC, Generic[T_InputRules]):
    """
    BaseImporter class which all importers inherit from.
    """

    @abstractmethod
    def to_rules(self) -> ReadRules[T_InputRules]:
        """Creates `Rules` object from the data for target role."""
        raise NotImplementedError()

    def _default_metadata(self) -> dict[str, Any]:
        creator = "UNKNOWN"
        with suppress(KeyError, ImportError):
            import getpass

            creator = getpass.getuser()

        return {
            "prefix": "neat",
            "schema": "partial",
            "namespace": Namespace("http://purl.org/cognite/neat/"),
            "version": "0.1.0",
            "title": "Neat Imported Data Model",
            "created": datetime.now().replace(microsecond=0).isoformat(),
            "updated": datetime.now().replace(microsecond=0).isoformat(),
            "creator": creator,
            "description": f"Imported using {type(self).__name__}",
        }

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls)


class _FutureResult:
    def __init__(self) -> None:
        self._result: Literal["success", "failure", "pending"] = "pending"

    @property
    def result(self) -> Literal["success", "failure", "pending"]:
        return self._result


@contextmanager
def _handle_issues(
    issues: IssueList,
    error_cls: type[NeatError] = NeatError,
    warning_cls: type[NeatWarning] = NeatWarning,
    error_args: dict[str, Any] | None = None,
) -> Iterator[_FutureResult]:
    """This is an internal help function to handle issues and warnings.

    Args:
        issues: The issues list to append to.
        error_cls: The class used to convert errors to issues.
        warning_cls:  The class used to convert warnings to issues.

    Returns:
        FutureResult: A future result object that can be used to check the result of the context manager.
    """
    with warnings.catch_warnings(record=True) as warning_logger:
        warnings.simplefilter("always")
        future_result = _FutureResult()
        try:
            yield future_result
        except ValidationError as e:
            issues.extend(error_cls.from_pydantic_errors(e.errors(), **(error_args or {})))
            future_result._result = "failure"
        else:
            future_result._result = "success"
        finally:
            if warning_logger:
                issues.extend([warning_cls.from_warning(warning) for warning in warning_logger])  # type: ignore[misc]



import sys
from collections.abc import Iterable
from typing import ClassVar, cast

from pydantic import BaseModel, ConfigDict, Field
from rdflib import Graph, Literal, Namespace
from rdflib.term import URIRef

from cognite.neat._constants import get_default_prefixes
from cognite.neat._rules.models._rdfpath import (
    Hop,
    Step,
)
from cognite.neat._utils.rdf_ import remove_namespace_from_uri, uri_to_short_form

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Triple(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    subject: str | URIRef
    predicate: str | URIRef
    object: str | URIRef | Literal | None = None
    optional: bool = Field(
        description="Indicates whether a triple is optional, used when building SPARQL query",
        default=False,
    )

    @classmethod
    def from_rdflib_triple(cls, triple: tuple[URIRef, URIRef, URIRef | Literal]) -> Self:
        return cls(subject=triple[0], predicate=triple[1], object=triple[2])


def generate_prefix_header(prefixes: dict[str, Namespace] | None = None) -> str:
    """Generate prefix header which is added to SPARQL query and allows for shorten query statements

    Parameters
    ----------
    prefixes : dict
        Dict containing prefix - namespace pairs, defaults to internal set of prefixes

    Returns
    -------
    str
        Prefix header
    """

    prefixes = prefixes or get_default_prefixes()

    return "".join(f"PREFIX {key}:<{value}>\n" for key, value in prefixes.items())


def get_predicate_id(
    graph: Graph,
    subject_type_id: str,
    object_type_id: str,
    prefixes: dict[str, Namespace] | None = None,
) -> URIRef:
    """Returns predicate (aka property) URI (i.e., ID) that connects subject and object

    Parameters
    ----------
    graph : Graph
        Data model graph or data model instance (aka knowledge graph)
    subject_type_id : str
        ID of subject type (aka subject class)
    object_type_id : str
        ID of object type (aka object class)
    prefixes : dict, optional
        Dict containing prefix - namespace pairs, defaults to internal set of prefixes

    Returns
    -------
    URIRef
        ID of predicate (aka property) connecting subject and object
    """

    prefixes = prefixes or get_default_prefixes()

    query = """

        SELECT ?predicateTypeID
        WHERE {
            ?subjectInstanceID a subjectTypeID .
            ?objectInstanceID a objectTypeID .
            ?subjectInstanceID ?predicateTypeID ?objectInstanceID .
        } LIMIT 1"""

    query = query.replace("insertPrefixes", generate_prefix_header(prefixes))
    final_query = query.replace("subjectTypeID", subject_type_id).replace("objectTypeID", object_type_id)
    res = list(cast(tuple, graph.query(final_query)))

    if len(res) != 1:
        raise ValueError("Subject and Object must have exactly 1 relation!")

    return res[0][0]


def hop2property_path(
    graph: Graph,
    hop: Hop,
    prefixes: dict[str, Namespace] | None = None,
) -> str:
    """Converts hop to property path string

    Parameters
    ----------
    graph : Graph
        Graph containing instances of classes
    hop : Hop
        Hop to convert
    prefixes : dict[str, Namespace]
        Dictionary of prefixes to use for compression and predicate querying

    Returns
    -------
    str
        Property path string for hop traversal (e.g. ^rdf:type/rdfs:subClassOf)
    """
    prefixes = prefixes if prefixes else get_default_prefixes()

    # setting previous step to origin, as we are starting from there
    previous_step = Step(class_=hop.class_, direction="origin")

    # add triples for all steps until destination
    property_path = ""
    for current_step in hop.traversal:
        sub_entity, obj_entity = (
            (current_step, previous_step) if current_step.direction == "source" else (previous_step, current_step)
        )

        predicate_raw = get_predicate_id(graph, sub_entity.class_.id, obj_entity.class_.id, prefixes)

        predicate = uri_to_short_form(predicate_raw, prefixes)

        predicate = f"^{predicate}" if current_step.direction == "source" else predicate
        property_path += f"{predicate}/"

        previous_step = current_step

    if previous_step.property:
        return property_path + previous_step.property.id
    else:
        # removing "/" at the end of property path if there is no property at the end
        return property_path[:-1]


def triples2dictionary(triples: Iterable[tuple[URIRef, URIRef, str | URIRef]]) -> dict[URIRef, dict[str, list[str]]]:
    """Converts list of triples to dictionary"""
    dictionary: dict[URIRef, dict[str, list[str]]] = {}
    for triple in triples:
        id_: str
        property_: str
        value: str
        uri: URIRef

        id_, property_, value = remove_namespace_from_uri(triple)  # type: ignore[misc]
        uri = triple[0]

        if uri not in dictionary:
            dictionary[uri] = {"external_id": [id_]}

        if property_ not in dictionary[uri]:
            dictionary[uri][property_] = [value]
        else:
            dictionary[uri][property_].append(value)
    return dictionary



from dataclasses import dataclass
from typing import Any

from rdflib import Namespace

from cognite.neat._rules.models._base_input import InputComponent, InputRules
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ClassEntity,
    MultiValueTypeInfo,
    UnknownEntity,
    load_value_type,
)
from cognite.neat._rules.models.information._rules_input import InformationInputClass, InformationInputMetadata

from ._rules import AssetClass, AssetMetadata, AssetProperty, AssetRules


@dataclass
class AssetInputMetadata(InformationInputMetadata):
    @classmethod
    def _get_verified_cls(cls) -> type[AssetMetadata]:
        return AssetMetadata


@dataclass
class AssetInputProperty(InputComponent[AssetProperty]):
    class_: ClassEntity | str
    property_: str
    value_type: DataType | ClassEntity | MultiValueTypeInfo | UnknownEntity | str
    name: str | None = None
    description: str | None = None
    comment: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None
    default: Any | None = None
    reference: str | None = None
    match_type: str | None = None
    transformation: str | None = None
    implementation: str | None = None
    # Only used internally
    inherited: bool = False

    @classmethod
    def _get_verified_cls(cls) -> type[AssetProperty]:
        return AssetProperty

    def dump(self, default_prefix: str) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["Class"] = ClassEntity.load(self.class_, prefix=default_prefix)
        output["Value Type"] = load_value_type(self.value_type, default_prefix)
        return output


@dataclass
class AssetInputClass(InformationInputClass):
    @classmethod
    def _get_verified_cls(cls) -> type[AssetClass]:
        return AssetClass


@dataclass
class AssetInputRules(InputRules[AssetRules]):
    metadata: AssetInputMetadata
    properties: list[AssetInputProperty]
    classes: list[AssetInputClass]
    prefixes: dict[str, Namespace] | None = None
    last: "AssetInputRules | None" = None
    reference: "AssetInputRules | None" = None

    @classmethod
    def _get_verified_cls(cls) -> type[AssetRules]:
        return AssetRules

    def dump(self) -> dict[str, Any]:
        default_prefix = self.metadata.prefix
        reference: dict[str, Any] | None = None
        if isinstance(self.reference, AssetInputRules):
            reference = self.reference.dump()
        elif isinstance(self.reference, AssetRules):
            # We need to load through the AssetRulesInput to set the correct default space and version
            reference = AssetInputRules.load(self.reference.model_dump()).dump()
        last: dict[str, Any] | None = None
        if isinstance(self.last, AssetInputRules):
            last = self.last.dump()
        elif isinstance(self.last, AssetRules):
            # We need to load through the AssetRulesInput to set the correct default space and version
            last = AssetInputRules.load(self.last.model_dump()).dump()

        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_prefix) for prop in self.properties],
            Classes=[class_.dump(default_prefix) for class_ in self.classes],
            Last=last,
            Reference=reference,
        )



from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Literal

import pandas as pd
from rdflib import Namespace

from cognite.neat._rules.models._base_input import InputComponent, InputRules
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ClassEntity,
    MultiValueTypeInfo,
    UnknownEntity,
    load_value_type,
)

from ._rules import (
    InformationClass,
    InformationMetadata,
    InformationProperty,
    InformationRules,
)


@dataclass
class InformationInputMetadata(InputComponent[InformationMetadata]):
    schema_: Literal["complete", "partial", "extended"]
    prefix: str
    namespace: str
    version: str
    creator: str
    data_model_type: Literal["solution", "enterprise"] = "enterprise"
    extension: Literal["addition", "reshape", "rebuild"] = "addition"
    name: str | None = None
    description: str | None = None
    created: datetime | str | None = None
    updated: datetime | str | None = None
    license: str | None = None
    rights: str | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationMetadata]:
        return InformationMetadata

    def dump(self, **kwargs) -> dict[str, Any]:
        output = super().dump()
        if self.created is None:
            output["created"] = datetime.now()
        if self.updated is None:
            output["updated"] = datetime.now()
        return output


@dataclass
class InformationInputProperty(InputComponent[InformationProperty]):
    class_: ClassEntity | str
    property_: str
    value_type: DataType | ClassEntity | MultiValueTypeInfo | UnknownEntity | str
    name: str | None = None
    description: str | None = None
    comment: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None
    default: Any | None = None
    reference: str | None = None
    match_type: str | None = None
    transformation: str | None = None
    # Only used internally
    inherited: bool = False

    @classmethod
    def _get_verified_cls(cls) -> type[InformationProperty]:
        return InformationProperty

    def dump(self, default_prefix: str, **kwargs) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["Class"] = ClassEntity.load(self.class_, prefix=default_prefix)
        output["Value Type"] = load_value_type(self.value_type, default_prefix)
        return output


@dataclass
class InformationInputClass(InputComponent[InformationClass]):
    class_: ClassEntity | str
    name: str | None = None
    description: str | None = None
    comment: str | None = None
    parent: str | list[ClassEntity] | None = None
    reference: str | None = None
    match_type: str | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationClass]:
        return InformationClass

    @property
    def class_str(self) -> str:
        return str(self.class_)

    def dump(self, default_prefix: str, **kwargs) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        parent: list[ClassEntity] | None = None
        if isinstance(self.parent, str):
            parent = [ClassEntity.load(parent, prefix=default_prefix) for parent in self.parent.split(",")]
        elif isinstance(self.parent, list):
            parent = [ClassEntity.load(parent_, prefix=default_prefix) for parent_ in self.parent]
        output["Class"] = ClassEntity.load(self.class_, prefix=default_prefix)
        output["Parent Class"] = parent
        return output


@dataclass
class InformationInputRules(InputRules[InformationRules]):
    metadata: InformationInputMetadata
    properties: list[InformationInputProperty] = field(default_factory=list)
    classes: list[InformationInputClass] = field(default_factory=list)
    prefixes: dict[str, Namespace] | None = None
    last: "InformationInputRules | None" = None
    reference: "InformationInputRules | None" = None

    @classmethod
    def _get_verified_cls(cls) -> type[InformationRules]:
        return InformationRules

    def dump(self) -> dict[str, Any]:
        default_prefix = self.metadata.prefix
        reference: dict[str, Any] | None = None
        if isinstance(self.reference, InformationInputRules):
            reference = self.reference.dump()
        elif isinstance(self.reference, InformationRules):
            # We need to load through the InformationRulesInput to set the correct default space and version
            reference = InformationInputRules.load(self.reference.model_dump()).dump()
        last: dict[str, Any] | None = None
        if isinstance(self.last, InformationInputRules):
            last = self.last.dump()
        elif isinstance(self.last, InformationRules):
            # We need to load through the InformationRulesInput to set the correct default space and version
            last = InformationInputRules.load(self.last.model_dump()).dump()

        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_prefix) for prop in self.properties],
            Classes=[class_.dump(default_prefix) for class_ in self.classes],
            Prefixes=self.prefixes,
            Last=last,
            Reference=reference,
        )

    def _repr_html_(self) -> str:
        summary = {
            "type": "Logical Data Model",
            "intended for": "Information Architect",
            "name": self.metadata.name,
            "external_id": self.metadata.prefix,
            "version": self.metadata.version,
            "classes": len(self.classes),
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

from cognite.neat._rules.importers._rdf._base import BaseRDFImporter
from cognite.neat._rules.models.data_types import _XSD_TYPES

from ._imf2classes import parse_imf_to_classes
from ._imf2metadata import parse_imf_metadata
from ._imf2properties import parse_imf_to_properties


class IMFImporter(BaseRDFImporter):
    """Convert SHACL shapes to tables/ transformation rules / Excel file.

        Args:
            filepath: Path to RDF file containing the SHACL Shapes

    !!! Note
        Rewrite to fit the SHACL rules we apply
        OWL Ontologies are information models which completeness varies. As such, constructing functional
        data model directly will often be impossible, therefore the produced Rules object will be ill formed.
        To avoid this, neat will automatically attempt to make the imported rules compliant by adding default
        values for missing information, attaching dangling properties to default containers based on the
        property type, etc.

        One has to be aware that NEAT will be opinionated about how to make the ontology
        compliant, and that the resulting rules may not be what you expect.

    """

    def _to_rules_components(
        self,
    ) -> dict:
        components = {
            "Metadata": parse_imf_metadata(),
            "Classes": parse_imf_to_classes(self.graph),
            "Properties": parse_imf_to_properties(self.graph),
        }

        return make_components_compliant(components)


def make_components_compliant(components: dict) -> dict:
    components = _add_missing_classes(components)
    components = _add_missing_value_types(components)
    components = _add_default_property_to_dangling_classes(components)

    return components


def _add_missing_classes(components: dict[str, list[dict]]) -> dict:
    """Add missing classes to Classes.

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing classes added to containers
    """

    missing_classes = {definition["Class"] for definition in components["Properties"]} - {
        definition["Class"] for definition in components["Classes"]
    }

    comment = (
        "Added by NEAT. "
        "This is a class that a domain of a property but was not defined in the ontology. "
        "It is added by NEAT to make the ontology compliant with CDF."
    )

    for class_ in missing_classes:
        components["Classes"].append({"Class": class_, "Comment": comment})

    return components


def _add_missing_value_types(components: dict) -> dict:
    """Add properties to classes that do not have any properties defined to them

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing properties added to containers
    """

    xsd_types = _XSD_TYPES
    candidate_value_types = {definition["Value Type"] for definition in components["Properties"]} - {
        definition["Class"] for definition in components["Classes"]
    }

    # to avoid issue of case sensitivity for xsd types
    value_types_lower = {v.lower() for v in candidate_value_types}

    xsd_types_lower = {x.lower() for x in xsd_types}

    # Create a mapping from lowercase strings to original strings
    value_types_mapping = {v.lower(): v for v in candidate_value_types}

    # Find the difference
    difference = value_types_lower - xsd_types_lower

    # Convert the difference back to the original case
    difference_original_case = {value_types_mapping[d] for d in difference}

    for class_ in difference_original_case:
        components["Classes"].append(
            {
                "Class": class_,
                "Comment": (
                    "Added by NEAT. "
                    "This is a class that a domain of a property but was not defined in the ontology. "
                    "It is added by NEAT to make the ontology compliant with CDF."
                ),
            }
        )

    return components


def _add_default_property_to_dangling_classes(components: dict[str, list[dict]]) -> dict:
    """Add missing classes to Classes.

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing classes added to containers
    """

    dangling_classes = {
        definition["Class"] for definition in components["Classes"] if not definition.get("Parent Class", None)
    } - {definition["Class"] for definition in components["Properties"]}

    comment = (
        "Added by NEAT. "
        "This is property has been added to this class since otherwise it will create "
        "dangling classes in the ontology."
    )

    for class_ in dangling_classes:
        components["Properties"].append(
            {
                "Class": class_,
                "Property": "label",
                "Value Type": "string",
                "Comment": comment,
                "Min Count": 0,
                "Max Count": 1,
                "Reference": "http://www.w3.org/2000/01/rdf-schema#label",
            }
        )

    return components



from collections import Counter
from collections.abc import Iterable, Sequence
from typing import TypeVar

T_Element = TypeVar("T_Element")


def remove_none_elements_from_set(s: set[T_Element]) -> set[T_Element]:
    return {x for x in s if x is not None}


def most_occurring_element(list_of_elements: list[T_Element]) -> T_Element:
    counts = Counter(list_of_elements)
    return counts.most_common(1)[0][0]


def chunker(sequence: Sequence[T_Element], chunk_size: int) -> Iterable[Sequence[T_Element]]:
    for i in range(0, len(sequence), chunk_size):
        yield sequence[i : i + chunk_size]


def remove_list_elements(input_list: list, elements_to_remove: list) -> list:
    return [element for element in input_list if element not in elements_to_remove]



from collections import Counter
from collections.abc import Collection, Sequence
from datetime import datetime
from pathlib import Path
from typing import Literal, cast

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier
from cognite.client.data_classes.data_modeling.containers import BTreeIndex, InvertedIndex
from cognite.client.data_classes.data_modeling.data_types import Enum as DMSEnum
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType, PropertyTypeWithUnit
from cognite.client.data_classes.data_modeling.views import (
    MultiEdgeConnectionApply,
    MultiReverseDirectRelationApply,
    SingleEdgeConnectionApply,
    SingleReverseDirectRelationApply,
    ViewPropertyApply,
)
from cognite.client.utils import ms_to_datetime

from cognite.neat._issues import IssueList, NeatIssue
from cognite.neat._issues.errors import FileTypeUnexpectedError, ResourceMissingIdentifierError, ResourceRetrievalError
from cognite.neat._issues.warnings import (
    PropertyNotFoundWarning,
    PropertyTypeNotSupportedWarning,
    ResourceNotFoundWarning,
    ResourcesDuplicatedWarning,
)
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.importers._base import BaseImporter, _handle_issues
from cognite.neat._rules.models import (
    DataModelType,
    DMSInputRules,
    DMSSchema,
    SchemaCompleteness,
)
from cognite.neat._rules.models.data_types import DataType, Enum
from cognite.neat._rules.models.dms import (
    DMSInputContainer,
    DMSInputEnum,
    DMSInputMetadata,
    DMSInputNode,
    DMSInputProperty,
    DMSInputView,
)
from cognite.neat._rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    EdgeEntity,
    ReverseConnectionEntity,
    ViewEntity,
)


class DMSImporter(BaseImporter[DMSInputRules]):
    """Imports a Data Model from Cognite Data Fusion.

    Args:
        schema: The schema containing the data model.
        read_issues: A list of issues that occurred during the import.
        metadata: Metadata for the data model.
        ref_metadata: Metadata for the reference data model.

    """

    def __init__(
        self,
        schema: DMSSchema,
        read_issues: Sequence[NeatIssue] | None = None,
        metadata: DMSInputMetadata | None = None,
        ref_metadata: DMSInputMetadata | None = None,
    ):
        # Calling this root schema to distinguish it from
        # * User Schema
        # * Reference Schema
        self.root_schema = schema
        self.metadata = metadata
        self.ref_metadata = ref_metadata
        self.issue_list = IssueList(read_issues)
        self._all_containers_by_id = schema.containers.copy()
        self._all_views_by_id = schema.views.copy()
        if schema.reference:
            self._all_containers_by_id.update(schema.reference.containers.items())
            self._all_views_by_id.update(schema.reference.views.items())

    @classmethod
    def from_data_model_id(
        cls,
        client: CogniteClient,
        data_model_id: DataModelIdentifier,
        reference_model_id: DataModelIdentifier | None = None,
    ) -> "DMSImporter":
        """Create a DMSImporter ready to convert the given data model to rules.

        Args:
            client: Instantiated CogniteClient to retrieve data model.
            reference_model_id: The reference data model to retrieve. This is the data model that
                the given data model is built on top of, typically, an enterprise data model.
            data_model_id: Data Model to retrieve.

        Returns:
            DMSImporter: DMSImporter instance
        """
        data_model_ids = [data_model_id, reference_model_id] if reference_model_id else [data_model_id]
        data_models = client.data_modeling.data_models.retrieve(data_model_ids, inline_views=True)

        user_models = cls._find_model_in_list(data_models, data_model_id)
        if len(user_models) == 0:
            return cls(
                DMSSchema(),
                [
                    ResourceRetrievalError(
                        dm.DataModelId.load(data_model_id),  # type: ignore[arg-type]
                        "data model",
                        "Data Model is missing in CDF",
                    )
                ],
            )
        user_model = user_models.latest_version()

        if reference_model_id:
            ref_models = cls._find_model_in_list(data_models, reference_model_id)
            if len(ref_models) == 0:
                return cls(
                    DMSSchema(),
                    [
                        ResourceRetrievalError(
                            dm.DataModelId.load(reference_model_id), "data model", "Data Model is missing in CDF"
                        )
                    ],
                )
            ref_model: dm.DataModel[dm.View] | None = ref_models.latest_version()
        else:
            ref_model = None

        issue_list = IssueList()
        with _handle_issues(issue_list) as result:
            schema = DMSSchema.from_data_model(client, user_model, ref_model)

        if result.result == "failure" or issue_list.has_errors:
            return cls(DMSSchema(), issue_list)

        metadata = cls._create_metadata_from_model(user_model, has_reference=ref_model is not None)
        ref_metadata = cls._create_metadata_from_model(ref_model) if ref_model else None

        return cls(schema, issue_list, metadata, ref_metadata)

    @classmethod
    def _find_model_in_list(
        cls, data_models: dm.DataModelList[dm.View], model_id: DataModelIdentifier
    ) -> dm.DataModelList[dm.View]:
        identifier = DataModelId.load(model_id)
        return dm.DataModelList[dm.View](
            [
                model
                for model in data_models
                if (model.space, model.external_id) == (identifier.space, identifier.external_id)
            ]
        )

    @classmethod
    def _create_metadata_from_model(
        cls,
        model: dm.DataModel[dm.View] | dm.DataModelApply,
        has_reference: bool = False,
    ) -> DMSInputMetadata:
        description, creator = DMSInputMetadata._get_description_and_creator(model.description)

        if isinstance(model, dm.DataModel):
            created = ms_to_datetime(model.created_time)
            updated = ms_to_datetime(model.last_updated_time)
        else:
            now = datetime.now().replace(microsecond=0)
            created = now
            updated = now
        return DMSInputMetadata(
            schema_="complete",
            data_model_type="solution" if has_reference else "enterprise",
            extension="addition",
            space=model.space,
            external_id=model.external_id,
            name=model.name or model.external_id,
            version=model.version or "0.1.0",
            updated=updated,
            created=created,
            creator=",".join(creator),
            description=description,
        )

    @classmethod
    def from_directory(cls, directory: str | Path) -> "DMSImporter":
        issue_list = IssueList()
        with _handle_issues(issue_list) as _:
            schema = DMSSchema.from_directory(directory)
        # If there were errors during the import, the to_rules
        return cls(schema, issue_list)

    @classmethod
    def from_zip_file(cls, zip_file: str | Path) -> "DMSImporter":
        if Path(zip_file).suffix != ".zip":
            return cls(DMSSchema(), [FileTypeUnexpectedError(Path(zip_file), frozenset([".zip"]))])
        issue_list = IssueList()
        with _handle_issues(issue_list) as _:
            schema = DMSSchema.from_zip(zip_file)
        return cls(schema, issue_list)

    def to_rules(self) -> ReadRules[DMSInputRules]:
        if self.issue_list.has_errors:
            # In case there were errors during the import, the to_rules method will return None
            return ReadRules(None, self.issue_list, {})

        if not self.root_schema.data_model:
            self.issue_list.append(ResourceMissingIdentifierError("data model", type(self.root_schema).__name__))
            return ReadRules(None, self.issue_list, {})

        model = self.root_schema.data_model

        schema_completeness = SchemaCompleteness.complete
        data_model_type = DataModelType.enterprise
        reference: DMSInputRules | None = None
        if (ref_schema := self.root_schema.reference) and (ref_model := ref_schema.data_model):
            # Reference should always be an enterprise model.
            reference = self._create_rule_components(
                ref_model,
                ref_schema,
                self.ref_metadata or self._create_default_metadata(list(ref_schema.views.values()), is_ref=True),
                DataModelType.enterprise,
            )
            data_model_type = DataModelType.solution

        user_rules = self._create_rule_components(
            model,
            self.root_schema,
            self.metadata,
            data_model_type,
            schema_completeness,
            has_reference=reference is not None,
        )
        user_rules.reference = reference

        return ReadRules(user_rules, self.issue_list, {})

    def _create_rule_components(
        self,
        data_model: dm.DataModelApply,
        schema: DMSSchema,
        metadata: DMSInputMetadata | None = None,
        data_model_type: DataModelType | None = None,
        schema_completeness: SchemaCompleteness | None = None,
        has_reference: bool = False,
    ) -> DMSInputRules:
        properties: list[DMSInputProperty] = []
        for view_id, view in schema.views.items():
            view_entity = ViewEntity.from_id(view_id)
            class_entity = view_entity.as_class()
            for prop_id, prop in (view.properties or {}).items():
                dms_property = self._create_dms_property(prop_id, prop, view_entity, class_entity)
                if dms_property is not None:
                    properties.append(dms_property)

        data_model_view_ids: set[dm.ViewId] = {
            view.as_id() if isinstance(view, dm.View | dm.ViewApply) else view for view in data_model.views or []
        }

        metadata = metadata or DMSInputMetadata.from_data_model(data_model, has_reference)
        if data_model_type is not None:
            metadata.data_model_type = str(data_model_type)  # type: ignore[assignment]
        if schema_completeness is not None:
            metadata.schema_ = str(schema_completeness)  # type: ignore[assignment]

        enum = self._create_enum_collections(schema.containers.values())

        return DMSInputRules(
            metadata=metadata,
            properties=properties,
            containers=[DMSInputContainer.from_container(container) for container in schema.containers.values()],
            views=[
                DMSInputView.from_view(view, in_model=view_id in data_model_view_ids)
                for view_id, view in schema.views.items()
            ],
            nodes=[DMSInputNode.from_node_type(node_type) for node_type in schema.node_types.values()],
            enum=enum,
        )

    @classmethod
    def _create_default_metadata(
        cls, views: Sequence[dm.View | dm.ViewApply], is_ref: bool = False
    ) -> DMSInputMetadata:
        now = datetime.now().replace(microsecond=0)
        space = Counter(view.space for view in views).most_common(1)[0][0]
        return DMSInputMetadata(
            schema_="complete",
            extension="addition",
            data_model_type="enterprise" if is_ref else "solution",
            space=space,
            external_id="Unknown",
            version="0.1.0",
            creator="Unknown",
            created=now,
            updated=now,
        )

    def _create_dms_property(
        self, prop_id: str, prop: ViewPropertyApply, view_entity: ViewEntity, class_entity: ClassEntity
    ) -> DMSInputProperty | None:
        if isinstance(prop, dm.MappedPropertyApply) and prop.container not in self._all_containers_by_id:
            self.issue_list.append(
                ResourceNotFoundWarning[dm.ContainerId, dm.PropertyId](
                    dm.ContainerId.load(prop.container),
                    "container",
                    view_entity.to_property_id(prop_id),
                    "view property",
                )
            )
            return None
        if (
            isinstance(prop, dm.MappedPropertyApply)
            and prop.container_property_identifier not in self._all_containers_by_id[prop.container].properties
        ):
            self.issue_list.append(
                PropertyNotFoundWarning(prop.container, "container", prop_id, view_entity.as_id(), "view"),
            )
            return None
        if not isinstance(
            prop,
            dm.MappedPropertyApply
            | SingleEdgeConnectionApply
            | MultiEdgeConnectionApply
            | SingleReverseDirectRelationApply
            | MultiReverseDirectRelationApply,
        ):
            self.issue_list.append(
                PropertyTypeNotSupportedWarning[dm.ViewId](view_entity.as_id(), "view", prop_id, type(prop).__name__)
            )
            return None

        value_type = self._get_value_type(prop, view_entity, prop_id)
        if value_type is None:
            return None

        return DMSInputProperty(
            class_=str(class_entity),
            property_=prop_id,
            description=prop.description,
            name=prop.name,
            connection=self._get_connection_type(prop),
            value_type=str(value_type),
            is_list=self._get_is_list(prop),
            nullable=self._get_nullable(prop),
            immutable=self._get_immutable(prop),
            default=self._get_default(prop),
            container=str(ContainerEntity.from_id(prop.container))
            if isinstance(prop, dm.MappedPropertyApply)
            else None,
            container_property=prop.container_property_identifier if isinstance(prop, dm.MappedPropertyApply) else None,
            view=str(view_entity),
            view_property=prop_id,
            index=self._get_index(prop, prop_id),
            constraint=self._get_constraint(prop, prop_id),
        )

    def _container_prop_unsafe(self, prop: dm.MappedPropertyApply) -> dm.ContainerProperty:
        """This method assumes you have already checked that the container with property exists."""
        return self._all_containers_by_id[prop.container].properties[prop.container_property_identifier]

    def _get_connection_type(
        self, prop: ViewPropertyApply
    ) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None:
        if isinstance(prop, SingleEdgeConnectionApply | MultiEdgeConnectionApply):
            properties = ViewEntity.from_id(prop.edge_source) if prop.edge_source is not None else None
            return EdgeEntity(
                properties=properties, type=DMSNodeEntity.from_reference(prop.type), direction=prop.direction
            )
        elif isinstance(prop, SingleReverseDirectRelationApply | MultiReverseDirectRelationApply):
            return ReverseConnectionEntity(property=prop.through.property)
        elif isinstance(prop, dm.MappedPropertyApply) and isinstance(
            self._container_prop_unsafe(prop).type, dm.DirectRelation
        ):
            return "direct"
        else:
            return None

    def _get_value_type(
        self, prop: ViewPropertyApply, view_entity: ViewEntity, prop_id
    ) -> DataType | ViewEntity | DMSUnknownEntity | None:
        if isinstance(
            prop,
            SingleEdgeConnectionApply
            | MultiEdgeConnectionApply
            | SingleReverseDirectRelationApply
            | MultiReverseDirectRelationApply,
        ):
            return ViewEntity.from_id(prop.source)
        elif isinstance(prop, dm.MappedPropertyApply):
            container_prop = self._container_prop_unsafe(cast(dm.MappedPropertyApply, prop))
            if isinstance(container_prop.type, dm.DirectRelation):
                if prop.source is None or prop.source not in self._all_views_by_id:
                    return DMSUnknownEntity()
                else:
                    return ViewEntity.from_id(prop.source)
            elif isinstance(container_prop.type, PropertyTypeWithUnit) and container_prop.type.unit:
                return DataType.load(f"{container_prop.type._type}(unit={container_prop.type.unit.external_id})")
            elif isinstance(container_prop.type, DMSEnum):
                return Enum(collection=ClassEntity(suffix=prop_id), unknownValue=container_prop.type.unknown_value)
            else:
                return DataType.load(container_prop.type._type)
        else:
            self.issue_list.append(
                PropertyTypeNotSupportedWarning[dm.ViewId](view_entity.as_id(), "view", prop_id, type(prop).__name__)
            )
            return None

    def _get_nullable(self, prop: ViewPropertyApply) -> bool | None:
        if isinstance(prop, dm.MappedPropertyApply):
            return self._container_prop_unsafe(prop).nullable
        else:
            return None

    def _get_immutable(self, prop: ViewPropertyApply) -> bool | None:
        if isinstance(prop, dm.MappedPropertyApply):
            return self._container_prop_unsafe(prop).immutable
        else:
            return None

    def _get_is_list(self, prop: ViewPropertyApply) -> bool | None:
        if isinstance(prop, dm.MappedPropertyApply):
            prop_type = self._container_prop_unsafe(prop).type
            return isinstance(prop_type, ListablePropertyType) and prop_type.is_list
        elif isinstance(prop, MultiEdgeConnectionApply | MultiReverseDirectRelationApply):
            return True
        elif isinstance(prop, SingleEdgeConnectionApply | SingleReverseDirectRelationApply):
            return False
        else:
            return None

    def _get_default(self, prop: ViewPropertyApply) -> str | None:
        if isinstance(prop, dm.MappedPropertyApply):
            default = self._container_prop_unsafe(prop).default_value
            if default is not None:
                return str(default)
        return None

    def _get_index(self, prop: ViewPropertyApply, prop_id) -> list[str] | None:
        if not isinstance(prop, dm.MappedPropertyApply):
            return None
        container = self._all_containers_by_id[prop.container]
        index: list[str] = []
        for index_name, index_obj in (container.indexes or {}).items():
            if isinstance(index_obj, BTreeIndex | InvertedIndex) and prop_id in index_obj.properties:
                index.append(index_name)
        return index or None

    def _get_constraint(self, prop: ViewPropertyApply, prop_id: str) -> list[str] | None:
        if not isinstance(prop, dm.MappedPropertyApply):
            return None
        container = self._all_containers_by_id[prop.container]
        unique_constraints: list[str] = []
        for constraint_name, constraint_obj in (container.constraints or {}).items():
            if isinstance(constraint_obj, dm.RequiresConstraint):
                # This is handled in the .from_container method of DMSContainer
                continue
            elif isinstance(constraint_obj, dm.UniquenessConstraint) and prop_id in constraint_obj.properties:
                unique_constraints.append(constraint_name)
            elif isinstance(constraint_obj, dm.UniquenessConstraint):
                # This does not apply to this property
                continue
            else:
                self.issue_list.append(
                    PropertyTypeNotSupportedWarning[dm.ContainerId](
                        prop.container, "container", prop_id, type(constraint_obj).__name__
                    )
                )
        return unique_constraints or None

    def _find_reverse_edge(
        self, prop_id: str, prop: SingleEdgeConnectionApply | MultiEdgeConnectionApply, view_id: dm.ViewId
    ) -> str | None:
        if prop.source not in self._all_views_by_id:
            return None
        view = self._all_views_by_id[prop.source]
        candidates = []
        for prop_name, reverse_prop in (view.properties or {}).items():
            if isinstance(reverse_prop, SingleEdgeConnectionApply | MultiEdgeConnectionApply):
                if (
                    reverse_prop.type == prop.type
                    and reverse_prop.source == view_id
                    and reverse_prop.direction != prop.direction
                ):
                    candidates.append(prop_name)
        if len(candidates) == 0:
            self.issue_list.append(
                PropertyNotFoundWarning(
                    prop.source,
                    "view property",
                    f"reverse edge of {prop_id}",
                    dm.PropertyId(view_id, prop_id),
                    "view property",
                )
            )
            return None
        if len(candidates) > 1:
            self.issue_list.append(
                ResourcesDuplicatedWarning(
                    frozenset(dm.PropertyId(view.as_id(), candidate) for candidate in candidates),
                    "view property",
                    default_action="Multiple reverse edges found for "
                    f"{dm.PropertyId(view_id, prop_id)!r}. Will use {candidates[0]}",
                )
            )

        return candidates[0]

    @staticmethod
    def _create_enum_collections(containers: Collection[dm.ContainerApply]) -> list[DMSInputEnum] | None:
        enum_collections: list[DMSInputEnum] = []
        for container in containers:
            for prop_id, prop in container.properties.items():
                if isinstance(prop.type, DMSEnum):
                    for identifier, value in prop.type.values.items():
                        enum_collections.append(
                            DMSInputEnum(
                                collection=prop_id, value=identifier, name=value.name, description=value.description
                            )
                        )
        return enum_collections



from cognite.client import data_modeling as dm

from cognite.neat._rules.transformers import SetIDDMSModel

from ._state import SessionState
from .exceptions import intercept_session_exceptions


@intercept_session_exceptions
class SetAPI:
    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def data_model_id(self, new_model_id: dm.DataModelId | tuple[str, str, str]) -> None:
        """Sets the data model ID of the latest verified data model."""
        if dms := self._state.last_verified_dms_rules:
            output = SetIDDMSModel(new_model_id).transform(dms)
            self._state.verified_rules.append(output.rules)
            if self._verbose:
                print(f"Data model ID set to {new_model_id}")
        else:
            print("No verified DMS data model available")



import colorsys
import random
from typing import Any, cast

import networkx as nx
from IPython.display import HTML, display
from pyvis.network import Network as PyVisNetwork  # type: ignore

from cognite.neat._constants import IN_NOTEBOOK, IN_PYODIDE
from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules.models.dms._rules import DMSRules
from cognite.neat._rules.models.entities._single_value import ClassEntity, ViewEntity
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._session.exceptions import NeatSessionError
from cognite.neat._utils.rdf_ import remove_namespace_from_uri

from ._state import SessionState
from .exceptions import intercept_session_exceptions


@intercept_session_exceptions
class ShowAPI:
    def __init__(self, state: SessionState) -> None:
        self._state = state
        self.data_model = ShowDataModelAPI(self._state)
        self.instances = ShowInstanceAPI(self._state)


@intercept_session_exceptions
class ShowBaseAPI:
    def __init__(self, state: SessionState) -> None:
        self._state = state

    def _generate_visualization(self, di_graph: nx.DiGraph, name: str) -> Any:
        if not IN_NOTEBOOK:
            raise NeatSessionError("Visualization is only available in Jupyter notebooks!")

        net = PyVisNetwork(
            notebook=IN_NOTEBOOK,
            cdn_resources="remote",
            directed=True,
            height="750px",
            width="100%",
            select_menu=IN_NOTEBOOK,
        )

        # Change the plotting layout
        net.repulsion(
            node_distance=100,
            central_gravity=0.3,
            spring_length=200,
            spring_strength=0.05,
            damping=0.09,
        )

        net.from_nx(di_graph)
        if IN_PYODIDE:
            net.write_html(name)
            return display(HTML(name))

        else:
            return net.show(name)


@intercept_session_exceptions
class ShowDataModelAPI(ShowBaseAPI):
    def __init__(self, state: SessionState) -> None:
        super().__init__(state)
        self._state = state

    def __call__(self) -> Any:
        if not self._state.has_verified_rules:
            raise NeatSessionError(
                "No verified data model available. Try using [bold].verify()[/bold] to verify data model."
            )

        try:
            di_graph = self._generate_dms_di_graph(self._state.last_verified_dms_rules)
            name = "dms_data_model.html"
        except NeatSessionError:
            di_graph = self._generate_info_di_graph(self._state.last_verified_information_rules)
            name = "information_data_model.html"

        return self._generate_visualization(di_graph, name)

    def _generate_dms_di_graph(self, rules: DMSRules) -> nx.DiGraph:
        """Generate a DiGraph from the last verified DMS rules."""
        di_graph = nx.DiGraph()

        # Add nodes and edges from Views sheet
        for view in rules.views:
            # if possible use human readable label coming from the view name
            if not di_graph.has_node(view.view.suffix):
                di_graph.add_node(view.view.suffix, label=view.name or view.view.suffix)

            # add implements as edges
            if view.implements:
                for implement in view.implements:
                    if not di_graph.has_node(implement.suffix):
                        di_graph.add_node(implement.suffix, label=implement.suffix)

                    di_graph.add_edge(
                        view.view.suffix,
                        implement.suffix,
                        label="implements",
                        dashes=True,
                    )

        # Add nodes and edges from Properties sheet
        for prop_ in rules.properties:
            if prop_.connection and isinstance(prop_.value_type, ViewEntity):
                if not di_graph.has_node(prop_.view.suffix):
                    di_graph.add_node(prop_.view.suffix, label=prop_.view.suffix)
                di_graph.add_edge(
                    prop_.view.suffix,
                    prop_.value_type.suffix,
                    label=prop_.name or prop_.property_,
                )

        return di_graph

    def _generate_info_di_graph(self, rules: InformationRules) -> nx.DiGraph:
        """Generate DiGraph representing information data model."""

        di_graph = nx.DiGraph()

        # Add nodes and edges from Views sheet
        for class_ in rules.classes:
            # if possible use human readable label coming from the view name
            if not di_graph.has_node(class_.class_.suffix):
                di_graph.add_node(
                    class_.class_.suffix,
                    label=class_.name or class_.class_.suffix,
                )

            # add subClassOff as edges
            if class_.parent:
                for parent in class_.parent:
                    if not di_graph.has_node(parent.suffix):
                        di_graph.add_node(parent.suffix, label=parent.suffix)
                    di_graph.add_edge(
                        class_.class_.suffix,
                        parent.suffix,
                        label="subClassOf",
                        dashes=True,
                    )

        # Add nodes and edges from Properties sheet
        for prop_ in rules.properties:
            if prop_.type_ == EntityTypes.object_property:
                if not di_graph.has_node(prop_.class_.suffix):
                    di_graph.add_node(prop_.class_.suffix, label=prop_.class_.suffix)

                di_graph.add_edge(
                    prop_.class_.suffix,
                    cast(ClassEntity, prop_.value_type).suffix,
                    label=prop_.name or prop_.property_,
                )

        return di_graph


@intercept_session_exceptions
class ShowInstanceAPI(ShowBaseAPI):
    def __init__(self, state: SessionState) -> None:
        super().__init__(state)
        self._state = state

    def __call__(self) -> Any:
        if not self._state.store.graph:
            raise NeatSessionError("No instances available. Try using [bold].read[/bold] to load instances.")

        di_graph = self._generate_instance_di_graph_and_types()
        return self._generate_visualization(di_graph, name="instances.html")

    def _generate_instance_di_graph_and_types(self) -> nx.DiGraph:
        query = """
        SELECT ?s ?p ?o ?ts ?to WHERE {
            ?s ?p ?o .
            FILTER(isIRI(?o))  # Example filter to check if ?o is an IRI (object type)
            FILTER(BOUND(?o))
            FILTER(?p != rdf:type)

            ?s a ?ts .
            ?o a ?to .
        }
        LIMIT 200
        """

        di_graph = nx.DiGraph()

        types = [type_ for type_, _ in self._state.store.queries.summarize_instances()]
        hex_colored_types = self._generate_hex_color_per_type(types)

        for (  # type: ignore
            subject,
            property_,
            object,
            subject_type,
            object_type,
        ) in self._state.store.graph.query(query):
            subject = remove_namespace_from_uri(subject)
            property_ = remove_namespace_from_uri(property_)
            object = remove_namespace_from_uri(object)
            subject_type = remove_namespace_from_uri(subject_type)
            object_type = remove_namespace_from_uri(object_type)

            di_graph.add_node(
                subject,
                label=subject,
                type=subject_type,
                title=subject_type,
                color=hex_colored_types[subject_type],
            )
            di_graph.add_node(
                object,
                label=object,
                type=object_type,
                title=object_type,
                color=hex_colored_types[object_type],
            )
            di_graph.add_edge(subject, object, label=property_, color="grey")

        return di_graph

    @staticmethod
    def _generate_hex_color_per_type(types: list[str]) -> dict:
        hex_colored_types = {}
        random.seed(381)
        for type_ in types:
            hue = random.random()
            saturation = random.uniform(0.5, 1.0)
            lightness = random.uniform(0.4, 0.6)
            rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
            hex_color = f"#{int(rgb[0] * 255):02x}{int(rgb[1] * 255):02x}{int(rgb[2] * 255):02x}"
            hex_colored_types[type_] = hex_color
        return hex_colored_types
